# -*- coding: utf-8 -*-
"""Implements the xonsh parser."""
import os
import sys
from collections import Iterable, Sequence, Mapping

from ply import yacc

from xonsh import ast
from xonsh.lexer import Lexer, LexToken
from xonsh.tools import (VER_3_4, VER_3_5, VER_3_5_1,
                         VER_MAJOR_MINOR, VER_FULL,
                         docstring_by_version)


class Location(object):
    """Location in a file."""

    def __init__(self, fname, lineno, column=None):
        """Takes a filename, line number, and optionally a column number."""
        self.fname = fname
        self.lineno = lineno
        self.column = column

    def __str__(self):
        s = '{0}:{1}'.format(self.fname, self.lineno)
        if self.column is not None:
            s += ':{0}'.format(self.column)
        return s


def has_elts(x):
    """Tests if x is an AST node with elements."""
    return isinstance(x, ast.AST) and hasattr(x, 'elts')


def ensure_has_elts(x, lineno=None, col_offset=None):
    """Ensures that x is an AST node with elements."""
    if not has_elts(x):
        if not isinstance(x, Iterable):
            x = [x]
        lineno = x[0].lineno if lineno is None else lineno
        col_offset = x[0].col_offset if col_offset is None else col_offset
        x = ast.Tuple(elts=x,
                      ctx=ast.Load(),
                      lineno=lineno,
                      col_offset=col_offset)
    return x


def empty_list(lineno=None, col=None):
    """Creates the AST node for an empty list."""
    return ast.List(elts=[], ctx=ast.Load(), lineno=lineno, col_offset=col)


def binop(x, op, y, lineno=None, col=None):
    """Creates the AST node for a binary operation."""
    lineno = x.lineno if lineno is None else lineno
    col = x.col_offset if col is None else col
    return ast.BinOp(left=x, op=op, right=y, lineno=lineno, col_offset=col)


def call_split_lines(x, lineno=None, col=None):
    """Creates the AST node for calling the 'splitlines' attribute of an
    object, nominally a string.
    """
    return ast.Call(func=ast.Attribute(value=x,
                                       attr='splitlines',
                                       ctx=ast.Load(),
                                       lineno=lineno,
                                       col_offset=col),
                    args=[],
                    keywords=[],
                    starargs=None,
                    kwargs=None,
                    lineno=lineno,
                    col_offset=col)


def ensure_list_from_str_or_list(x, lineno=None, col=None):
    """Creates the AST node for the following expression::

        [x] if isinstance(x, str) else x

    Somewhat useful.
    """
    return ast.IfExp(test=ast.Call(func=ast.Name(id='isinstance',
                                                 ctx=ast.Load(),
                                                 lineno=lineno,
                                                 col_offset=col),
                                   args=[x, ast.Name(id='str',
                                                     ctx=ast.Load(),
                                                     lineno=lineno,
                                                     col_offset=col)],
                                   keywords=[],
                                   starargs=None,
                                   kwargs=None,
                                   lineno=lineno,
                                   col_offset=col),
                     body=ast.List(elts=[x],
                                   ctx=ast.Load(),
                                   lineno=lineno,
                                   col_offset=col),
                     orelse=x,
                     lineno=lineno,
                     col_offset=col)


def xonsh_call(name, args, lineno=None, col=None):
    """Creates the AST node for calling a function of a given name."""
    return ast.Call(func=ast.Name(id=name,
                                  ctx=ast.Load(),
                                  lineno=lineno,
                                  col_offset=col),
                    args=args,
                    keywords=[],
                    starargs=None,
                    kwargs=None,
                    lineno=lineno,
                    col_offset=col)


def xonsh_help(x, lineno=None, col=None):
    """Creates the AST node for calling the __xonsh_help__() function."""
    return xonsh_call('__xonsh_help__', [x], lineno=lineno, col=col)


def xonsh_superhelp(x, lineno=None, col=None):
    """Creates the AST node for calling the __xonsh_superhelp__() function."""
    return xonsh_call('__xonsh_superhelp__', [x], lineno=lineno, col=col)


def xonsh_regexpath(x, lineno=None, col=None):
    """Creates the AST node for calling the __xonsh_regexpath__() function."""
    return xonsh_call('__xonsh_regexpath__', [x], lineno=lineno, col=col)


def load_ctx(x):
    """Recursively sets ctx to ast.Load()"""
    if not hasattr(x, 'ctx'):
        return
    x.ctx = ast.Load()
    if isinstance(x, (ast.Tuple, ast.List)):
        for e in x.elts:
            load_ctx(e)
    elif isinstance(x, ast.Starred):
        load_ctx(x.value)


def store_ctx(x):
    """Recursively sets ctx to ast.Store()"""
    if not hasattr(x, 'ctx'):
        return
    x.ctx = ast.Store()
    if isinstance(x, (ast.Tuple, ast.List)):
        for e in x.elts:
            store_ctx(e)
    elif isinstance(x, ast.Starred):
        store_ctx(x.value)


def empty_list_if_newline(x):
    return [] if x == '\n' else x


def lopen_loc(x):
    """Extracts the line and column number for a node that may have anb opening
    parenthesis, brace, or braket.
    """
    lineno = x._lopen_lineno if hasattr(x, '_lopen_lineno') else x.lineno
    col = x._lopen_col if hasattr(x, '_lopen_col') else x.col_offset
    return lineno, col


class Parser(object):
    """A class that parses the xonsh language."""

    def __init__(self,
                 lexer_optimize=True,
                 lexer_table='xonsh.lexer_table',
                 yacc_optimize=True,
                 yacc_table='xonsh.parser_table',
                 yacc_debug=False,
                 outputdir=None):
        """Parameters
        ----------
        lexer_optimize : bool, optional
            Set to false when unstable and true when lexer is stable.
        lexer_table : str, optional
            Lexer module used when optimized.
        yacc_optimize : bool, optional
            Set to false when unstable and true when parser is stable.
        yacc_table : str, optional
            Parser module used when optimized.
        yacc_debug : debug, optional
            Dumps extra debug info.
        outputdir : str or None, optional
            The directory to place generated tables within.
        """
        self.lexer = lexer = Lexer()
        self.tokens = lexer.tokens

        opt_rules = [
            'newlines', 'arglist', 'func_call', 'rarrow_test', 'typedargslist',
            'equals_test', 'colon_test', 'tfpdef', 'comma_tfpdef_list',
            'comma_pow_tfpdef', 'vfpdef', 'comma_vfpdef_list',
            'comma_pow_vfpdef', 'equals_yield_expr_or_testlist_list',
            'testlist', 'as_name', 'period_or_ellipsis_list',
            'comma_import_as_name_list', 'comma_dotted_as_name_list',
            'comma_name_list', 'comma_test', 'elif_part_list', 'finally_part',
            'varargslist', 'or_and_test_list', 'and_not_test_list',
            'comp_op_expr_list', 'xor_and_expr_list',
            'ampersand_shift_expr_list', 'shift_arith_expr_list',
            'op_factor_list', 'trailer_list', 'testlist_comp',
            'yield_expr_or_testlist_comp', 'dictorsetmaker',
            'comma_subscript_list', 'test', 'sliceop', 'comp_iter',
            'yield_arg', 'test_comma_list',]
        if VER_MAJOR_MINOR <= VER_3_4:
            opt_rules += ['argument_comma_list', 'comma_argument_list',]
        for rule in opt_rules:
            self._opt_rule(rule)

        list_rules = [
            'comma_tfpdef', 'comma_vfpdef', 'semi_small_stmt',
            'comma_test_or_star_expr', 'period_or_ellipsis',
            'comma_import_as_name', 'comma_dotted_as_name', 'period_name',
            'comma_name', 'elif_part', 'except_part', 'comma_with_item',
            'or_and_test', 'and_not_test', 'comp_op_expr', 'pipe_xor_expr',
            'xor_and_expr', 'ampersand_shift_expr', 'shift_arith_expr',
            'pm_term', 'op_factor', 'trailer', 'comma_subscript',
            'comma_expr_or_star_expr', 'comma_test', 'comma_argument', 'comma_item',
            'attr_period_name', 'test_comma', 'equals_yield_expr_or_testlist',
            'test_or_star_expr']
        if VER_MAJOR_MINOR <= VER_3_4:
            list_rules += ['argument_comma',]
        for rule in list_rules:
            self._list_rule(rule)

        tok_rules = ['def', 'class', 'async', 'return', 'number', 'name', 
                     'none', 'true', 'false', 'ellipsis', 'if', 'del', 'assert', 
                     'lparen', 'lbrace', 'lbracket', 'string', 'times', 'plus', 
                     'minus', 'divide', 'doublediv', 'mod', 'at', 'lshift', 'rshift',
                     'pipe', 'xor', 'ampersand', 'await', 'for', 'colon',
                     'import', 'except', 'nonlocal', 'global', 'yield', 'from', 
                     'raise', 'with', 'dollar_lparen', 'dollar_lbrace', 
                     'dollar_lbracket', 'try']
        for rule in tok_rules:
            self._tok_rule(rule)

        yacc_kwargs = dict(module=self,
                           debug=yacc_debug,
                           start='start_symbols',
                           optimize=yacc_optimize,
                           tabmodule=yacc_table)
        if not yacc_debug:
            yacc_kwargs['errorlog'] = yacc.NullLogger()
        if outputdir is not None:
            yacc_kwargs['outputdir'] = outputdir
        self.parser = yacc.yacc(**yacc_kwargs)

        # Keeps track of the last token given to yacc (the lookahead token)
        self._last_yielded_token = None

    def reset(self):
        """Resets for clean parsing."""
        self.lexer.reset()
        self._last_yielded_token = None

    def parse(self, s, filename='<code>', mode='exec', debug_level=0):
        """Returns an abstract syntax tree of xonsh code.

        Parameters
        ----------
        s : str
            The xonsh code.
        filename : str, optional
            Name of the file.
        mode : str, optional
            Execution mode, one of: exec, eval, or single.
        debug_level : str, optional
            Debugging level passed down to yacc.

        Returns
        -------
        tree : AST
        """
        self.reset()
        self.lexer.fname = filename
        tree = self.parser.parse(input=s, lexer=self.lexer, debug=debug_level)
        # hack for getting modes right
        if mode == 'single':
            if isinstance(tree, ast.Expression):
                tree = ast.Interactive(body=[self.expr(tree.body)])
            elif isinstance(tree, ast.Module):
                tree = ast.Interactive(body=tree.body)
        return tree

    def _lexer_errfunc(self, msg, line, column):
        self._parse_error(msg, self.currloc(line, column))

    def _yacc_lookahead_token(self):
        """Gets the next-to-last and last token seen by the lexer."""
        return self.lexer.beforelast, self.lexer.last

    def _opt_rule(self, rulename):
        """For a rule name, creates an associated optional rule.
        '_opt' is appended to the rule name.
        """

        def optfunc(self, p):
            p[0] = p[1]

        optfunc.__doc__ = ('{0}_opt : empty\n'
                           '        | {0}').format(rulename)
        optfunc.__name__ = 'p_' + rulename + '_opt'
        setattr(self.__class__, optfunc.__name__, optfunc)

    def _list_rule(self, rulename):
        """For a rule name, creates an associated list rule.
        '_list' is appended to the rule name.
        """

        def listfunc(self, p):
            p[0] = p[1] if len(p) == 2 else p[1] + p[2]

        listfunc.__doc__ = ('{0}_list : {0}\n'
                            '         | {0}_list {0}').format(rulename)
        listfunc.__name__ = 'p_' + rulename + '_list'
        setattr(self.__class__, listfunc.__name__, listfunc)

    def _tok_rule(self, rulename):
        """For a rule name, creates a rule that retuns the corresponding token.
        '_tok' is appended to the rule name.
        """

        def tokfunc(self, p):
            s, t = self._yacc_lookahead_token()
            uprule = rulename.upper()
            if s is not None and s.type == uprule:
                p[0] = s
            elif t is not None and t.type == uprule:
                p[0] = t
            else:
                raise TypeError('token for {0!r} not found.'.format(rulename))

        tokfunc.__doc__ = '{0}_tok : {1}'.format(rulename, rulename.upper())
        tokfunc.__name__ = 'p_' + rulename + '_tok'
        setattr(self.__class__, tokfunc.__name__, tokfunc)

    def currloc(self, lineno, column=None):
        """Returns the current location."""
        return Location(fname=self.lexer.fname, lineno=lineno, column=column)

    def expr(self, p):
        """Creates an expression for a token."""
        expr = ast.Expr(value=p, lineno=p.lineno, col_offset=p.col_offset)
        expr.max_lineno = self.lineno
        expr.max_col = self.col
        return expr

    def token_col(self, t):
        """Gets ths token column"""
        return t.lexpos

    @property
    def lineno(self):
        if self.lexer.last is None:
            return 1
        else:
            return self.lexer.last.lineno

    @property
    def col(self):
        s, t = self._yacc_lookahead_token()
        if t is not None:
            #raise ValueError
            if t.type == 'NEWLINE':
                t = s
            return self.token_col(t)
        return 0

    def _parse_error(self, msg, loc):
        err = SyntaxError('{0}: {1}'.format(loc, msg))
        err.loc = loc
        raise err

    #
    # Precedence of operators
    #
    precedence = (('left', 'PIPE'), ('left', 'XOR'), ('left', 'AMPERSAND'),
                  ('left', 'EQ', 'NE'), ('left', 'GT', 'GE', 'LT', 'LE'),
                  ('left', 'RSHIFT', 'LSHIFT'), ('left', 'PLUS', 'MINUS'),
                  ('left', 'TIMES', 'DIVIDE', 'DOUBLEDIV', 'MOD'),
                  ('left', 'POW'), )

    #
    # Grammar as defined by BNF
    #

    def p_start_symbols(self, p):
        """start_symbols : single_input
                         | file_input
                         | eval_input
                         | empty
        """
        p[0] = p[1]

    def p_single_input(self, p):
        """single_input : compound_stmt NEWLINE
        """
        p1 = empty_list_if_newline(p[1])
        p0 = ast.Interactive(body=p1)
        p[0] = p0

    def p_file_input(self, p):
        """file_input : file_stmts"""
        p[0] = ast.Module(body=p[1])

    def p_file_stmts(self, p):
        """file_stmts : newline_or_stmt
                      | file_stmts newline_or_stmt
        """
        if len(p) == 2:
            # newline_or_stmt ENDMARKER
            p1 = empty_list_if_newline(p[1])
            p[0] = p1
        else:
            # file_input newline_or_stmt ENDMARKER
            p2 = empty_list_if_newline(p[2])
            p[0] = p[1] + p2

    def p_newline_or_stmt(self, p):
        """newline_or_stmt : NEWLINE
                           | stmt
        """
        p[0] = p[1]

    def p_newlines(self, p):
        """newlines : NEWLINE
                    | newlines NEWLINE
        """
        p[0] = p[1] if len(p) == 2 else p[1] + p[2]

    def p_eval_input(self, p):
        """eval_input : testlist newlines_opt
        """
        p[0] = ast.Expression(body=p[1])

    def p_func_call(self, p):
        """func_call : LPAREN arglist_opt RPAREN"""
        p[0] = p[2]

    def p_attr_period_name(self, p):
        """attr_period_name : PERIOD NAME"""
        p[0] = [p[2]]

    def p_attr_name(self, p):
        """attr_name : name_tok
                     | name_tok attr_period_name_list
        """
        p1 = p[1]
        name = ast.Name(id=p1.value,
                        ctx=ast.Load(),
                        lineno=p1.lineno,
                        col_offset=p1.lexpos)
        if len(p) == 2:
            p0 = name
        else:
            p2 = p[2]
            p0 = ast.Attribute(value=name,
                               attr=p2[0],
                               ctx=ast.Load(),
                               lineno=p1.lineno,
                               col_offset=p1.lexpos)
            for a in p2[1:]:
                p0 = ast.Attribute(value=p0,
                                   attr=a,
                                   ctx=ast.Load(),
                                   lineno=p0.lineno,
                                   col_offset=p0.col_offset)
        p[0] = p0

    def p_decorator(self, p):
        """decorator : at_tok attr_name NEWLINE
                     | at_tok attr_name func_call NEWLINE
        """
        lenp = len(p)
        p1, name = p[1], p[2]
        p3 = p[3] if lenp > 3 else None
        if isinstance(name, ast.Attribute) or (lenp == 5 and p3 is not None):
            lineno, col = name.lineno, name.col_offset
        else:
            lineno, col = p1.lineno, p1.lexpos
        if lenp == 4:
            p0 = name
        elif p3 is None:
            p0 = ast.Call(func=name,
                          args=[],
                          keywords=[],
                          starargs=None,
                          kwargs=None,
                          lineno=lineno,
                          col_offset=col)
        else:
            p0 = ast.Call(func=name,
                          lineno=lineno,
                          col_offset=col, **p3)
        p[0] = p0

    def p_decorators(self, p):
        """decorators : decorator
                      | decorators decorator
        """
        p[0] = [p[1]] if len(p) == 2 else p[1] + [p[2]]

    @docstring_by_version(
        v34=\
        """classdef_or_funcdef : classdef
                               | funcdef
        """,
        v35=\
        """classdef_or_funcdef : classdef
                               | funcdef
                               | async_funcdef
        """,
        )
    def p_classdef_or_funcdef(self, p):
        p[0] = p[1]

    def p_decorated(self, p):
        """decorated : decorators classdef_or_funcdef"""
        p1, p2 = p[1], p[2]
        targ = p2[0]
        targ.decorator_list = p1
        # this is silly, CPython. This claims a func or class starts on 
        # the line of the first decorator, rather than the 'def' or 'class'
        # line.  However, it retains the original col_offset.
        targ.lineno = p1[0].lineno
        # async functions take the col number of the 'def', unless they are
        # decorated, in which case they have the col of the 'async'. WAT?
        if hasattr(targ, '_async_tok'):
            targ.col_offset = targ._async_tok.lexpos
            del targ._async_tok
        p[0] = p2

    def p_rarrow_test(self, p):
        """rarrow_test : RARROW test"""
        p[0] = p[2]

    def p_funcdef(self, p):
        """funcdef : def_tok NAME parameters rarrow_test_opt COLON suite"""
        f = ast.FunctionDef(name=p[2],
                            args=p[3],
                            returns=p[4],
                            body=p[6],
                            decorator_list=[],
                            lineno=p[1].lineno,
                            col_offset=p[1].lexpos)
        p[0] = [f]

    def p_async_funcdef(self, p):
        """async_funcdef : async_tok funcdef"""
        p1, f = p[1], p[2][0]
        p[0] = [ast.AsyncFunctionDef(**f.__dict__)]
        p[0][0]._async_tok = p1

    def p_parameters(self, p):
        """parameters : LPAREN typedargslist_opt RPAREN"""
        p2 = p[2]
        if p2 is None:
            p2 = ast.arguments(args=[],
                               vararg=None,
                               kwonlyargs=[],
                               kw_defaults=[],
                               kwarg=None,
                               defaults=[])
        p[0] = p2

    def p_equals_test(self, p):
        """equals_test : EQUALS test"""
        p[0] = p[2]

    def p_typedargslist(self, p):
        """typedargslist : tfpdef equals_test_opt comma_tfpdef_list_opt comma_opt
                         | tfpdef equals_test_opt comma_tfpdef_list_opt comma_opt TIMES tfpdef_opt COMMA POW vfpdef
                         | tfpdef equals_test_opt comma_tfpdef_list_opt comma_opt TIMES tfpdef_opt comma_tfpdef_list_opt
                         | tfpdef equals_test_opt comma_tfpdef_list_opt comma_opt TIMES tfpdef_opt comma_tfpdef_list COMMA POW tfpdef
                         | tfpdef equals_test_opt comma_tfpdef_list_opt comma_opt POW tfpdef
                         | TIMES tfpdef_opt comma_tfpdef_list comma_pow_tfpdef_opt
                         | TIMES tfpdef_opt comma_pow_tfpdef_opt
                         | POW tfpdef
        """
        lenp = len(p)
        p1, p2 = p[1], p[2]
        p3 = p[3] if lenp > 3 else None
        p4 = p[4] if lenp > 4 else None
        # skip p5
        p6 = p[6] if lenp > 6 else None
        p7 = p[7] if lenp > 7 else None
        # skip p8
        p9 = p[9] if lenp > 9 else None
        p10 = p[10] if lenp > 10 else None
        p0 = ast.arguments(args=[],
                           vararg=None,
                           kwonlyargs=[],
                           kw_defaults=[],
                           kwarg=None,
                           defaults=[])
        if lenp == 3:
            p0.kwarg = p2
        elif lenp == 4:
            self._set_var_args(p0, p2, None)
            p0.kwarg = p3
        elif lenp == 5 and p1 != '*':
            # x
            self._set_regular_args(p0, p1, p2, p3, p4)
        elif lenp == 5 and p1 == '*':
            self._set_var_args(p0, p2, p3)  # *args
            if p4 is not None:
                # *args, x, **kwargs
                p0.kwarg = p4
        elif lenp == 7:
            # x, **kwargs
            self._set_regular_args(p0, p1, p2, p3, p4)
            p0.kwarg = p[6]
        elif lenp == 8:
            # x, *args
            self._set_regular_args(p0, p1, p2, p3, p4)
            self._set_var_args(p0, p6, p7)
        elif lenp == 10:
            # x, *args, **kwargs
            self._set_regular_args(p0, p1, p2, p3, p4)
            self._set_var_args(p0, p6, None)
            p0.kwarg = p9
        elif lenp == 11:
            # x, *args, **kwargs
            self._set_regular_args(p0, p1, p2, p3, p4)
            self._set_var_args(p0, p6, p7)
            p0.kwarg = p10
        else:
            assert False
        p[0] = p0

    def p_colon_test(self, p):
        """colon_test : COLON test"""
        p[0] = p[2]

    def p_tfpdef(self, p):
        """tfpdef : name_tok colon_test_opt"""
        p1 = p[1]
        kwargs = {'arg': p1.value,
                  'annotation': p[2]}
        if VER_FULL >= VER_3_5_1:
            kwargs.update({
                'lineno': p1.lineno,
                'col_offset': p1.lexpos,
            })
        p[0] = ast.arg(**kwargs)

    def p_comma_tfpdef(self, p):
        """comma_tfpdef : COMMA
                        | COMMA tfpdef equals_test_opt
        """
        if len(p) == 2:
            p[0] = []
        else:
            p[0] = [{'arg': p[2], 'default': p[3]}]

    def p_comma_pow_tfpdef(self, p):
        """comma_pow_tfpdef : COMMA POW tfpdef"""
        p[0] = p[3]

    def _set_args_def(self, argmts, vals, kwargs=False):
        args, defs = (argmts.kwonlyargs, argmts.kw_defaults) if kwargs else \
                     (argmts.args, argmts.defaults)
        for v in vals:
            args.append(v['arg'])
            d = v['default']
            if kwargs or (d is not None):
                defs.append(d)

    def _set_regular_args(self, p0, p1, p2, p3, p4):
        if p2 is None and p3 is None:
            # x
            p0.args.append(p1)
        elif p2 is not None and p3 is None:
            # x=42
            p0.args.append(p1)
            p0.defaults.append(p2)
        elif p2 is None and p3 is not None:
            # x, y and x, y=42
            p0.args.append(p1)
            self._set_args_def(p0, p3)
        else:
            # x=42, y=42
            p0.args.append(p1)
            p0.defaults.append(p2)
            self._set_args_def(p0, p3)

    def _set_var_args(self, p0, vararg, kwargs):
        if vararg is None:
            self._set_args_def(p0, kwargs, kwargs=True)
        elif vararg is not None and kwargs is None:
            # *args
            p0.vararg = vararg
        elif vararg is not None and kwargs is not None:
            # *args, x and *args, x, y and *args, x=10 and *args, x=10, y
            # and *args, x, y=10, and *args, x=42, y=65
            p0.vararg = vararg
            self._set_args_def(p0, kwargs, kwargs=True)
        else:
            assert False

    def p_varargslist(self, p):
        """varargslist : vfpdef equals_test_opt comma_vfpdef_list_opt comma_opt
                       | vfpdef equals_test_opt comma_vfpdef_list_opt comma_opt TIMES vfpdef_opt COMMA POW vfpdef
                       | vfpdef equals_test_opt comma_vfpdef_list_opt comma_opt TIMES vfpdef_opt comma_vfpdef_list_opt
                       | vfpdef equals_test_opt comma_vfpdef_list_opt comma_opt TIMES vfpdef_opt comma_vfpdef_list COMMA POW vfpdef
                       | vfpdef equals_test_opt comma_vfpdef_list_opt comma_opt POW vfpdef
                       | TIMES vfpdef_opt comma_vfpdef_list comma_pow_vfpdef_opt
                       | TIMES vfpdef_opt comma_pow_vfpdef_opt
                       | POW vfpdef
        """
        lenp = len(p)
        p1, p2 = p[1], p[2]
        p3 = p[3] if lenp > 3 else None
        p4 = p[4] if lenp > 4 else None
        # skip p5
        p6 = p[6] if lenp > 6 else None
        p7 = p[7] if lenp > 7 else None
        # skip p8
        p9 = p[9] if lenp > 9 else None
        p10 = p[10] if lenp > 10 else None
        p0 = ast.arguments(args=[],
                           vararg=None,
                           kwonlyargs=[],
                           kw_defaults=[],
                           kwarg=None,
                           defaults=[])
        if lenp == 3:
            p0.kwarg = p2
        elif lenp == 4:
            self._set_var_args(p0, p2, None)
            p0.kwarg = p3
        elif lenp == 5 and p1 != '*':
            # x
            self._set_regular_args(p0, p1, p2, p3, p4)
        elif lenp == 5 and p1 == '*':
            self._set_var_args(p0, p2, p3)  # *args
            if p4 is not None:
                # *args, x, **kwargs
                p0.kwarg = p4
        elif lenp == 7:
            # x, **kwargs
            self._set_regular_args(p0, p1, p2, p3, p4)
            p0.kwarg = p6
        elif lenp == 8:
            # x, *args
            self._set_regular_args(p0, p1, p2, p3, p4)
            self._set_var_args(p0, p6, p7)
        elif lenp == 10:
            # x, *args, **kwargs
            self._set_regular_args(p0, p1, p2, p3, p4)
            self._set_var_args(p0, p6, None)
            p0.kwarg = p9
        elif lenp == 11:
            # x, *args, **kwargs
            self._set_regular_args(p0, p1, p2, p3, p4)
            self._set_var_args(p0, p6, p7)
            p0.kwarg = p10
        else:
            assert False
        p[0] = p0

    def p_vfpdef(self, p):
        """vfpdef : name_tok"""
        p1 = p[1]
        kwargs = {'arg': p1.value,
                  'annotation': None}
        if VER_FULL >= VER_3_5_1:
            kwargs.update({
                'lineno': p1.lineno,
                'col_offset': p1.lexpos,
            })
        p[0] = ast.arg(**kwargs)

    def p_comma_vfpdef(self, p):
        """comma_vfpdef : COMMA
                        | COMMA vfpdef equals_test_opt
        """
        if len(p) == 2:
            p[0] = []
        else:
            p[0] = [{'arg': p[2], 'default': p[3]}]

    def p_comma_pow_vfpdef(self, p):
        """comma_pow_vfpdef : COMMA POW vfpdef"""
        p[0] = p[3]

    def p_stmt(self, p):
        """stmt : simple_stmt
                | compound_stmt
        """
        p[0] = p[1]

    def p_stmt_list(self, p):
        """stmt_list : stmt
                     | stmt_list stmt"""
        if len(p) == 2:
            p[0] = p[1]
        else:
            p[0] = p[1] + p[2]

    def p_semi_opt(self, p):
        """semi_opt : SEMI
                    | empty
        """
        if len(p) == 2:
            p[0] = p[1]

    def p_semi_small_stmt(self, p):
        """semi_small_stmt : SEMI small_stmt"""
        p[0] = [p[2]]

    def p_simple_stmt(self, p):
        """simple_stmt : small_stmt semi_small_stmt_list semi_opt NEWLINE
                       | small_stmt semi_opt NEWLINE
        """
        p1, p2 = p[1], p[2]
        p0 = [p1]
        if p2 is not None and p2 != ';':
            p0 += p2
        p[0] = p0

    def p_small_stmt(self, p):
        """small_stmt : expr_stmt
                      | del_stmt
                      | pass_stmt
                      | flow_stmt
                      | import_stmt
                      | global_stmt
                      | nonlocal_stmt
                      | assert_stmt
        """
        p[0] = p[1]

    _augassign_op = {
        '+=': ast.Add,
        '-=': ast.Sub,
        '*=': ast.Mult,
        '@=': ast.MatMult,
        '/=': ast.Div,
        '%=': ast.Mod,
        '//=': ast.FloorDiv,
        '**=': ast.Pow,
        '^=': ast.BitXor,
        '&=': ast.BitAnd,
        '|=': ast.BitOr,
        '<<=': ast.LShift,
        '>>=': ast.RShift
    }

    def p_expr_stmt(self, p):
        """expr_stmt : testlist_star_expr augassign yield_expr_or_testlist
                     | testlist_star_expr equals_yield_expr_or_testlist_list_opt
                     | testlist equals_yield_expr_or_testlist_list_opt
                     | test_comma_list_opt star_expr comma_test_list equals_yield_expr_or_testlist
                     | test_comma_list_opt star_expr comma_opt test_comma_list_opt equals_yield_expr_or_testlist
        """
        lenp = len(p)
        p1, p2 = p[1], p[2]
        p1 = [] if p1 is None else p1
        if isinstance(p1, ast.Tuple):
            p1 = [p1]
        for targ in p1:
            store_ctx(targ)
        if lenp == 3:
            if p2 is None and len(p1) == 1:
                load_ctx(p1[0])
                p0 = self.expr(p1[0])
            elif p2 is None:
                assert False
            else:
                list(map(store_ctx, p2[:-1]))
                lineno, col = lopen_loc(p1[0])
                p0 = ast.Assign(targets=p1 + p2[:-1], value=p2[-1],
                                lineno=lineno, col_offset=col)
        elif lenp == 4:
            op = self._augassign_op[p2]
            if op is None:
                self._parse_error('operation {0!r} not supported'.format(p2),
                                  self.currloc(lineno=p.lineno, column=p.lexpos))
            p0 = ast.AugAssign(target=p1[0], op=op(), value=p[3],
                               lineno=p1[0].lineno, col_offset=p1[0].col_offset)
        elif lenp == 5 or lenp == 6:
            if lenp == 5:
                targs, rhs = p[3], p[4][0]
            else:
                targs, rhs = (p[4] or []), p[5][0]
            store_ctx(p2)
            for targ in targs:
                store_ctx(targ)
            p1.append(p2)
            p1.extend(targs)
            p1 = [ast.Tuple(elts=p1,
                            ctx=ast.Store(),
                            lineno=p1[0].lineno,
                            col_offset=p1[0].col_offset)]
            p0 = ast.Assign(targets=p1,
                            value=rhs,
                            lineno=p1[0].lineno,
                            col_offset=p1[0].col_offset)
        else:
            assert False
        p[0] = p0

    def p_test_comma(self, p):
        """test_comma : test COMMA"""
        p[0] = [p[1]]

    def p_comma_opt(self, p):
        """comma_opt : COMMA
                     | empty
        """
        if len(p) == 2:
            p[0] = p[1]

    def p_test_or_star_expr(self, p):
        """test_or_star_expr : test
                             | star_expr
        """
        p[0] = p[1]

    def p_comma_test_or_star_expr(self, p):
        """comma_test_or_star_expr : COMMA test_or_star_expr"""
        p[0] = [p[2]]

    def p_testlist_star_expr(self, p):
        """testlist_star_expr : test_or_star_expr comma_test_or_star_expr_list comma_opt
                              | test_or_star_expr comma_opt
        """
        p1, p2 = p[1], p[2]
        if p2 is None:
            p0 = [p1]
        elif p2 == ',':
            p0 = [ast.Tuple(elts=[p1],
                            ctx=ast.Load(),
                            lineno=p1.lineno,
                            col_offset=p1.col_offset)]
        else:
            p0 = [ast.Tuple(elts=[p1] + p2,
                            ctx=ast.Load(),
                            lineno=p1.lineno,
                            col_offset=p1.col_offset)]
        p[0] = p0

    def p_augassign(self, p):
        """augassign : PLUSEQUAL
                     | MINUSEQUAL
                     | TIMESEQUAL
                     | ATEQUAL
                     | DIVEQUAL
                     | MODEQUAL
                     | AMPERSANDEQUAL
                     | PIPEEQUAL
                     | XOREQUAL
                     | LSHIFTEQUAL
                     | RSHIFTEQUAL
                     | POWEQUAL
                     | DOUBLEDIVEQUAL
        """
        p[0] = p[1]

    def p_yield_expr_or_testlist(self, p):
        """yield_expr_or_testlist : yield_expr
                                  | testlist
        """
        p[0] = p[1]

    def p_equals_yield_expr_or_testlist(self, p):
        """equals_yield_expr_or_testlist : EQUALS yield_expr_or_testlist"""
        p[0] = [p[2]]

    #
    # For normal assignments, additional restrictions enforced
    # by the interpreter
    #
    def p_del_stmt(self, p):
        """del_stmt : del_tok exprlist"""
        p1 = p[1]
        p2 = p[2]
        for targ in p2:
            targ.ctx = ast.Del()
        p0 = ast.Delete(targets=p2, lineno=p1.lineno, col_offset=p1.lexpos)
        p[0] = p0

    def p_pass_stmt(self, p):
        """pass_stmt : PASS"""
        p[0] = ast.Pass(lineno=self.lineno, col_offset=self.col)

    def p_flow_stmt(self, p):
        """flow_stmt : break_stmt
                     | continue_stmt
                     | return_stmt
                     | raise_stmt
                     | yield_stmt
        """
        p[0] = p[1]

    def p_break_stmt(self, p):
        """break_stmt : BREAK"""
        p[0] = ast.Break(lineno=self.lineno, col_offset=self.col)

    def p_continue_stmt(self, p):
        """continue_stmt : CONTINUE"""
        p[0] = ast.Continue(lineno=self.lineno, col_offset=self.col)

    def p_return_stmt(self, p):
        """return_stmt : return_tok testlist_opt"""
        p1 = p[1]
        p[0] = ast.Return(value=p[2], lineno=p1.lineno, col_offset=p1.lexpos)

    def p_yield_stmt(self, p):
        """yield_stmt : yield_expr"""
        p[0] = self.expr(p[1])

    def p_raise_stmt(self, p):
        """raise_stmt : raise_tok
                      | raise_tok test
                      | raise_tok test FROM test
        """
        p1 = p[1]
        lenp = len(p)
        cause = None
        if lenp == 2:
            exc = None
        elif lenp == 3:
            exc = p[2]
        elif lenp == 5:
            exc = p[2]
            cause = p[4]
        else:
            assert False
        p0 = ast.Raise(exc=exc,
                       cause=cause,
                       lineno=p1.lineno,
                       col_offset=p1.lexpos)
        p[0] = p0

    def p_import_stmt(self, p):
        """import_stmt : import_name
                       | import_from
        """
        p[0] = p[1]

    def p_import_name(self, p):
        """import_name : import_tok dotted_as_names
        """
        p1 = p[1]
        p[0] = ast.Import(names=p[2], lineno=p1.lineno, col_offset=p1.lexpos)

    def p_import_from_pre(self, p):
        """import_from_pre : from_tok period_or_ellipsis_list_opt dotted_name
                           | from_tok period_or_ellipsis_list
        """
        p1 = p[1]
        if len(p) == 3:
            p0 = p[2]
        elif len(p) == 4:
            p2, p3 = p[2], p[3]
            p0 = p3 if p2 is None else p2 + p3
        else:
            assert False
        p[0] = (p0, p1.lineno, p1.lexpos)

    def p_import_from_post(self, p):
        """import_from_post : TIMES
                            | LPAREN import_as_names RPAREN
                            | import_as_names
        """
        if len(p) == 2:
            if p[1] == '*':
                p[1] = [ast.alias(name='*', asname=None)]
            p0 = p[1]
        elif len(p) == 4:
            p0 = p[2]
        else:
            assert False
        p[0] = p0

    def p_import_from(self, p):
        """import_from : import_from_pre IMPORT import_from_post
        """
        # note below: the ('.' | '...') is necessary because '...' is
        # tokenized as ELLIPSIS
        p1, lineno, col = p[1]
        mod = p1.lstrip('.')
        lvl = len(p1) - len(mod)
        mod = mod or None
        p[0] = ast.ImportFrom(module=mod,
                              names=p[3],
                              level=lvl,
                              lineno=lineno,
                              col_offset=col)

    def p_period_or_ellipsis(self, p):
        """period_or_ellipsis : PERIOD
                              | ELLIPSIS
        """
        p[0] = p[1]

    def p_as_name(self, p):
        """as_name : AS NAME"""
        p[0] = p[2]

    def p_import_as_name(self, p):
        """import_as_name : NAME as_name_opt"""
        p[0] = ast.alias(name=p[1], asname=p[2])

    def p_comma_import_as_name(self, p):
        """comma_import_as_name : COMMA import_as_name
        """
        p[0] = [p[2]]

    def p_dotted_as_name(self, p):
        """dotted_as_name : dotted_name as_name_opt"""
        p0 = ast.alias(name=p[1], asname=p[2])
        p[0] = p0

    def p_comma_dotted_as_name(self, p):
        """comma_dotted_as_name : COMMA dotted_as_name"""
        p[0] = [p[2]]

    def p_import_as_names(self, p):
        """import_as_names : import_as_name comma_import_as_name_list_opt comma_opt
        """
        p1, p2 = p[1], p[2]
        p0 = [p1]
        if p2 is not None:
            p0.extend(p2)
        p[0] = p0

    def p_dotted_as_names(self, p):
        """dotted_as_names : dotted_as_name comma_dotted_as_name_list_opt"""
        p1, p2 = p[1], p[2]
        p0 = [p1]
        if p2 is not None:
            p0.extend(p2)
        p[0] = p0

    def p_period_name(self, p):
        """period_name : PERIOD NAME"""
        p[0] = p[1] + p[2]

    def p_dotted_name(self, p):
        """dotted_name : NAME
                       | NAME period_name_list
        """
        p[0] = p[1] if len(p) == 2 else p[1] + p[2]

    def p_comma_name(self, p):
        """comma_name : COMMA NAME"""
        p[0] = [p[2]]

    def p_global_stmt(self, p):
        """global_stmt : global_tok NAME comma_name_list_opt"""
        p1, p2, p3 = p[1], p[2], p[3]
        names = [p2]
        if p3 is not None:
            names += p3
        p[0] = ast.Global(names=names, lineno=p1.lineno, col_offset=p1.lexpos)

    def p_nonlocal_stmt(self, p):
        """nonlocal_stmt : nonlocal_tok NAME comma_name_list_opt"""
        p1, p2, p3 = p[1], p[2], p[3]
        names = [p2]
        if p3 is not None:
            names += p3
        p[0] = ast.Nonlocal(names=names,
                            lineno=p1.lineno,
                            col_offset=p1.lexpos)

    def p_comma_test(self, p):
        """comma_test : COMMA test"""
        p[0] = [p[2]]

    def p_assert_stmt(self, p):
        """assert_stmt : assert_tok test comma_test_opt"""
        p1, p2, p3 = p[1], p[2], p[3]
        if p3 is not None:
            if len(p3) != 1:
                assert False
            p3 = p3[0]
        p0 = ast.Assert(test=p2,
                        msg=p3,
                        lineno=p1.lineno,
                        col_offset=p1.lexpos)
        p[0] = p0

    def p_compound_stmt(self, p):
        """compound_stmt : if_stmt
                         | while_stmt
                         | for_stmt
                         | try_stmt
                         | with_stmt
                         | funcdef
                         | classdef
                         | decorated
                         | async_stmt
        """
        p[0] = p[1]

    def p_elif_part(self, p):
        """elif_part : ELIF test COLON suite"""
        p2 = p[2]
        p[0] = [ast.If(test=p2,
                       body=p[4],
                       orelse=[],
                       lineno=p2.lineno,
                       col_offset=p2.col_offset)]

    def p_else_part(self, p):
        """else_part : ELSE COLON suite"""
        p[0] = p[3]

    def p_if_stmt(self, p):
        """if_stmt : if_tok test COLON suite elif_part_list_opt
                   | if_tok test COLON suite elif_part_list_opt else_part
        """
        p1 = p[1]
        lastif = ast.If(test=p[2],
                        body=p[4],
                        orelse=[],
                        lineno=p1.lineno,
                        col_offset=p1.lexpos)
        p0 = [lastif]
        p5 = p[5]
        p6 = p[6] if len(p) > 6 else []
        if p5 is not None:
            for elseif in p5:
                lastif.orelse.append(elseif)
                lastif = elseif
        lastif.orelse = p6
        p[0] = p0

    def p_while_stmt(self, p):
        """while_stmt : WHILE test COLON suite
                      | WHILE test COLON suite else_part
        """
        p5 = p[5] if len(p) > 5 else []
        p[0] = [ast.While(test=p[2],
                          body=p[4],
                          orelse=p5,
                          lineno=self.lineno,
                          col_offset=self.col)]

    def p_for_stmt(self, p):
        """for_stmt : for_tok exprlist IN testlist COLON suite
                    | for_tok exprlist IN testlist COLON suite else_part
        """
        p1, p2 = p[1], p[2]
        p7 = p[7] if len(p) > 7 else []
        if len(p2) == 1:
            p2 = p2[0]
            store_ctx(p2)
        else:
            for x in p2:
                store_ctx(x)
            p2 = ast.Tuple(elts=p2,
                           ctx=ast.Store(),
                           lineno=p2[0].lineno,
                           col_offset=p2[0].col_offset)
        p[0] = [ast.For(target=p2,
                        iter=p[4],
                        body=p[6],
                        orelse=p7,
                        lineno=p1.lineno,
                        col_offset=p1.lexpos)]

    def p_async_for_stmt(self, p):
        """async_for_stmt : ASYNC for_stmt"""
        f = p[2][0]
        p[0] = [ast.AsyncFor(**f.__dict__)]

    def p_except_part(self, p):
        """except_part : except_clause COLON suite"""
        p0 = p[1]
        p0.body = p[3]
        p[0] = [p0]

    def p_finally_part(self, p):
        """finally_part : FINALLY COLON suite"""
        p[0] = p[3]

    def p_try_stmt(self, p):
        """try_stmt : try_tok COLON suite except_part_list else_part finally_part_opt
                    | try_tok COLON suite except_part_list finally_part_opt
                    | try_tok COLON suite finally_part
        """
        lenp = len(p)
        p1 = p[1]
        t = ast.Try(body=p[3], lineno=p1.lineno, col_offset=p1.lexpos)
        if lenp == 7:
            p5, p6 = p[5], p[6]
            t.handlers = p[4]
            t.orelse = [] if p5 is None else p5
            t.finalbody = [] if p6 is None else p6
        elif lenp == 6:
            p5 = p[5]
            t.handlers = p[4]
            t.orelse = []
            t.finalbody = [] if p5 is None else p5
        else:
            t.handlers = []
            t.orelse = []
            t.finalbody = p[4]
        p[0] = [t]

    def p_with_stmt(self, p):
        """with_stmt : with_tok with_item COLON suite
                     | with_tok with_item comma_with_item_list COLON suite
        """
        p1, p2, p3 = p[1], [p[2]], p[3]
        if len(p) == 5:
            body = p[4]
        else:
            p2 += p3
            body = p[5]
        p[0] = [ast.With(items=p2,
                         body=body,
                         lineno=p1.lineno,
                         col_offset=p1.lexpos)]

    def p_async_with_stmt(self, p):
        """async_with_stmt : ASYNC with_stmt"""
        w = p[2][0]
        p[0] = [ast.AsyncWith(**w.__dict__)]

    def p_as_expr(self, p):
        """as_expr : AS expr"""
        p2 = p[2]
        store_ctx(p2)
        p[0] = p2

    def p_with_item(self, p):
        """with_item : test
                     | test as_expr
        """
        p2 = p[2] if len(p) > 2 else None
        p[0] = ast.withitem(context_expr=p[1], optional_vars=p2)

    def p_comma_with_item(self, p):
        """comma_with_item : COMMA with_item"""
        p[0] = [p[2]]

    def p_except_clause(self, p):
        """except_clause : except_tok
                         | except_tok test as_name_opt
        """
        p1 = p[1]
        if len(p) == 2:
            p0 = ast.ExceptHandler(type=None,
                                   name=None,
                                   lineno=p1.lineno,
                                   col_offset=p1.lexpos)
        else:
            p0 = ast.ExceptHandler(type=p[2],
                                   name=p[3],
                                   lineno=p1.lineno,
                                   col_offset=p1.lexpos)
        p[0] = p0

    def p_async_stmt(self, p):
        """async_stmt : async_funcdef
                      | async_with_stmt
                      | async_for_stmt
        """
        p[0] = p[1]

    def p_suite(self, p):
        """suite : simple_stmt
                 | NEWLINE INDENT stmt_list DEDENT
        """
        p[0] = p[1] if len(p) == 2 else p[3]

    def p_test(self, p):
        """test : or_test
                | or_test IF or_test ELSE test
                | lambdef
        """
        if len(p) == 2:
            p0 = p[1]
        else:
            p0 = ast.IfExp(test=p[3],
                           body=p[1],
                           orelse=p[5],
                           lineno=self.lineno,
                           col_offset=self.col)
        p[0] = p0

    def p_test_nocond(self, p):
        """test_nocond : or_test
                       | lambdef_nocond
        """
        p[0] = p[1]

    def p_lambdef(self, p):
        """lambdef : LAMBDA varargslist_opt COLON test"""
        p2, p4 = p[2], p[4]
        if p2 is None:
            args = ast.arguments(args=[],
                                 vararg=None,
                                 kwonlyargs=[],
                                 kw_defaults=[],
                                 kwarg=None,
                                 defaults=[])
        else:
            args = p2
        p0 = ast.Lambda(args=args,
                        body=p4,
                        lineno=self.lineno,
                        col_offset=self.col)
        p[0] = p0

    def p_lambdef_nocond(self, p):
        """lambdef_nocond : LAMBDA varargslist_opt COLON test_nocond"""
        assert False

    def p_or_test(self, p):
        """or_test : and_test or_and_test_list_opt"""
        p1, p2 = p[1], p[2]
        if p2 is None:
            p0 = p1
        elif len(p2) == 2:
            lineno, col = lopen_loc(p1)
            p0 = ast.BoolOp(op=p2[0],
                            values=[p1, p2[1]],
                            lineno=lineno,
                            col_offset=col)
        else:
            lineno, col = lopen_loc(p1)
            p0 = ast.BoolOp(op=p2[0],
                            values=[p[1]] + p2[1::2],
                            lineno=lineno,
                            col_offset=col)
        p[0] = p0

    def p_or_and_test(self, p):
        """or_and_test : OR and_test"""
        p[0] = [ast.Or(), p[2]]

    def p_and_test(self, p):
        """and_test : not_test and_not_test_list_opt"""
        p1, p2 = p[1], p[2]
        if p2 is None:
            p0 = p1
        elif len(p2) == 2:
            lineno, col = lopen_loc(p1)
            p0 = ast.BoolOp(op=p2[0],
                            values=[p1, p2[1]],
                            lineno=lineno,
                            col_offset=col)
        else:
            lineno, col = lopen_loc(p1)
            p0 = ast.BoolOp(op=p2[0],
                            values=[p1] + p2[1::2],
                            lineno=lineno,
                            col_offset=col)
        p[0] = p0

    def p_and_not_test(self, p):
        """and_not_test : AND not_test"""
        p[0] = [ast.And(), p[2]]

    def p_not_test(self, p):
        """not_test : NOT not_test
                    | comparison
        """
        if len(p) == 2:
            p0 = p[1]
        else:
            p0 = ast.UnaryOp(op=ast.Not(),
                             operand=p[2],
                             lineno=self.lineno,
                             col_offset=self.col)
        p[0] = p0

    def p_comparison(self, p):
        """comparison : expr comp_op_expr_list_opt"""
        p1 = p[1]
        p2 = p[2]
        if p2 is None:
            p0 = p1
        else:
            p0 = ast.Compare(left=p1,
                             ops=p2[::2],
                             comparators=p2[1::2],
                             lineno=p1.lineno,
                             col_offset=p1.col_offset)
        p[0] = p0

    def p_comp_op_expr(self, p):
        """comp_op_expr : comp_op expr"""
        p[0] = [p[1], p[2]]

    _comp_ops = {
        '<': ast.Lt,
        '>': ast.Gt,
        '==': ast.Eq,
        '>=': ast.GtE,
        '<=': ast.LtE,
        '!=': ast.NotEq,
        'in': ast.In,
        ('not', 'in'): ast.NotIn,
        'is': ast.Is,
        ('is', 'not'): ast.IsNot
    }

    def p_comp_op(self, p):
        """comp_op : LT
                   | GT
                   | EQ
                   | GE
                   | LE
                   | NE
                   | IN
                   | NOT IN
                   | IS
                   | IS NOT
        """
        key = p[1] if len(p) == 2 else (p[1], p[2])
        p[0] = self._comp_ops[key]()

    def p_star_expr(self, p):
        """star_expr : times_tok expr"""
        p1 = p[1]
        p[0] = ast.Starred(value=p[2],
                           ctx=ast.Load(),
                           lineno=p1.lineno,
                           col_offset=p1.lexpos)

    def _binop_combine(self, p1, p2):
        """Combines binary operations"""
        if p2 is None:
            p0 = p1
        elif isinstance(p2, ast.BinOp):
            p2.left = p1
            p0 = p2
        elif isinstance(p2, Sequence) and isinstance(p2[0], ast.BinOp):
            p0 = p2[0]
            p0.left = p1
            p0.lineno, p0.col_offset = lopen_loc(p1)
            for bop in p2[1:]:
                locer = p1 if p0.left is p1 else p0
                bop.left = p0
                p0.lineno, p0.col_offset = lopen_loc(locer)
                p0 = bop
        else:
            p0 = p1 + p2
        return p0

    def p_expr(self, p):
        """expr : xor_expr
                | xor_expr pipe_xor_expr_list
        """
        p[0] = self._binop_combine(p[1], p[2] if len(p) > 2 else None)

    def p_pipe_xor_expr(self, p):
        """pipe_xor_expr : pipe_tok xor_expr"""
        p1 = p[1]
        p[0] = [ast.BinOp(left=None,
                          op=ast.BitOr(),
                          right=p[2],
                          lineno=p1.lineno,
                          col_offset=p1.lexpos)]

    def p_xor_expr(self, p):
        """xor_expr : and_expr xor_and_expr_list_opt"""
        p[0] = self._binop_combine(p[1], p[2])

    def p_xor_and_expr(self, p):
        """xor_and_expr : xor_tok and_expr"""
        p1 = p[1]
        p[0] = [ast.BinOp(left=None,
                          op=ast.BitXor(),
                          right=p[2],
                          lineno=p1.lineno,
                          col_offset=p1.lexpos)]

    def p_and_expr(self, p):
        """and_expr : shift_expr ampersand_shift_expr_list_opt"""
        p[0] = self._binop_combine(p[1], p[2])

    def p_ampersand_shift_expr(self, p):
        """ampersand_shift_expr : ampersand_tok shift_expr"""
        p1 = p[1]
        p[0] = [ast.BinOp(left=None,
                          op=ast.BitAnd(),
                          right=p[2],
                          lineno=p1.lineno,
                          col_offset=p1.lexpos)]

    def p_shift_expr(self, p):
        """shift_expr : arith_expr shift_arith_expr_list_opt"""
        p[0] = self._binop_combine(p[1], p[2])

    def p_shift_arith_expr(self, p):
        """shift_arith_expr : lshift_tok arith_expr
                            | rshift_tok arith_expr
        """
        p1 = p[1]
        op = ast.LShift() if p1.value == '<<' else ast.RShift()
        p[0] = [ast.BinOp(left=None,
                          op=op,
                          right=p[2],
                          lineno=p1.lineno,
                          col_offset=p1.lexpos)]

    def p_arith_expr(self, p):
        """arith_expr : term
                      | term pm_term_list
        """
        p1 = p[1]
        p2 = p[2] if len(p) > 2 else None
        if p2 is None:
            p0 = p1
        elif len(p2) == 2:
            lineno, col = lopen_loc(p1)
            p0 = ast.BinOp(left=p1,
                           op=p2[0],
                           right=p2[1],
                           lineno=lineno,
                           col_offset=col)
        else:
            left = p1
            for op, right in zip(p2[::2], p2[1::2]):
                locer = left if left is p1 else op
                lineno, col = lopen_loc(locer)
                left = ast.BinOp(left=left,
                                 op=op,
                                 right=right,
                                 lineno=lineno,
                                 col_offset=col)
            p0 = left
        p[0] = p0

    _term_binops = {
        '+': ast.Add,
        '-': ast.Sub,
        '*': ast.Mult,
        '@': ast.MatMult,
        '/': ast.Div,
        '%': ast.Mod,
        '//': ast.FloorDiv
    }

    def p_pm_term(self, p):
        """pm_term : plus_tok term
                   | minus_tok term
        """
        p1 = p[1]
        op = self._term_binops[p1.value](lineno=p1.lineno, col_offset=p1.lexpos)
        p[0] = [op, p[2]]

    def p_term(self, p):
        """term : factor op_factor_list_opt"""
        p1, p2 = p[1], p[2]
        if p2 is None:
            p0 = p1
        elif len(p2) == 2:
            lineno, col = lopen_loc(p1)
            p0 = ast.BinOp(left=p1,
                           op=p2[0],
                           right=p2[1],
                           lineno=lineno,
                           col_offset=col)
        else:
            left = p1
            for op, right in zip(p2[::2], p2[1::2]):
                locer = left if left is p1 else op
                lineno, col = lopen_loc(locer)
                left = ast.BinOp(left=left,
                                 op=op,
                                 right=right,
                                 lineno=lineno,
                                 col_offset=col)
            p0 = left
        p[0] = p0

    def p_op_factor(self, p):
        """op_factor : times_tok factor
                     | at_tok factor
                     | divide_tok factor
                     | mod_tok factor
                     | doublediv_tok factor
        """
        p1 = p[1]
        op = self._term_binops[p1.value]
        if op is None:
            self._parse_error('operation {0!r} not supported'.format(p1),
                              self.currloc(lineno=p.lineno, column=p.lexpos))
        p[0] = [op(lineno=p1.lineno, col_offset=p1.lexpos), p[2]]

    _factor_ops = {'+': ast.UAdd, '-': ast.USub, '~': ast.Invert}

    def p_factor(self, p):
        """factor : PLUS factor
                  | MINUS factor
                  | TILDE factor
                  | power
        """
        if len(p) == 2:
            p0 = p[1]
        else:
            op = self._factor_ops[p[1]]()
            p0 = ast.UnaryOp(op=op,
                             operand=p[2],
                             lineno=self.lineno,
                             col_offset=self.col)
        p[0] = p0

    def p_power(self, p):
        """power : atom_expr
                 | atom_expr POW factor
        """
        lenp = len(p)
        p1 = p[1]
        if lenp == 2:
            p0 = p1
        elif lenp == 4:
            # actual power rule
            p0 = ast.BinOp(left=p1,
                           op=ast.Pow(),
                           right=p[3],
                           lineno=p1.lineno,
                           col_offset=p1.col_offset)
        else:
            assert False
        p[0] = p0

    def p_yield_expr_or_testlist_comp(self, p):
        """yield_expr_or_testlist_comp : yield_expr
                                       | testlist_comp
        """
        p[0] = p[1]

    def _list_or_elts_if_not_real_tuple(self, x):
        if isinstance(x, ast.Tuple) and not (hasattr(x, '_real_tuple') and \
                                             x._real_tuple):
            rtn = x.elts
        else:
            rtn = [x]
        return rtn

    @docstring_by_version(
        v34="""atom_expr : atom trailer_list_opt""",
        v35=\
        """atom_expr : atom trailer_list_opt
                     | await_tok atom trailer_list_opt
        """
        )
    def p_atom_expr(self, p):
        lenp = len(p)
        if lenp == 3:
            leader, trailers = p[1], p[2]
        elif lenp == 4:
            leader, trailers = p[2], p[3]
        else:
            assert False
        p0 = leader
        if trailers is None:
            trailers = []
        for trailer in trailers:
            if isinstance(trailer, (ast.Index, ast.Slice)):
                p0 = ast.Subscript(value=leader,
                                   slice=trailer,
                                   ctx=ast.Load(),
                                   lineno=leader.lineno,
                                   col_offset=leader.col_offset)
            elif isinstance(trailer, Mapping):
                p0 = ast.Call(func=leader,
                              lineno=leader.lineno,
                              col_offset=leader.col_offset, **trailer)
            elif isinstance(trailer, str):
                if trailer == '?':
                    p0 = xonsh_help(leader, lineno=leader.lineno, col=leader.col_offset)
                elif trailer == '??':
                    p0 = xonsh_superhelp(leader,
                                         lineno=leader.lineno,
                                         col=leader.col_offset)
                else:
                    p0 = ast.Attribute(value=leader,
                                       attr=trailer,
                                       ctx=ast.Load(),
                                       lineno=leader.lineno,
                                       col_offset=leader.col_offset)
            else:
                assert False
            leader = p0
        if lenp == 4:
            p1 = p[1]
            p0 = ast.Await(value=p0, ctx=ast.Load(), lineno=p1.lineno,
                           col_offset=p1.lexpos)
        p[0] = p0

    def p_atom(self, p):
        """atom : lparen_tok yield_expr_or_testlist_comp_opt RPAREN
                | lbracket_tok testlist_comp_opt RBRACKET
                | lbrace_tok dictorsetmaker_opt RBRACE
                | name_tok
                | number
                | string_literal_list
                | ellipsis_tok
                | none_tok
                | true_tok
                | false_tok
                | REGEXPATH
                | DOLLAR_NAME
                | dollar_lbrace_tok test RBRACE
                | dollar_lparen_tok subproc RPAREN
                | dollar_lbracket_tok subproc RBRACKET
        """
        p1 = p[1]
        if len(p) == 2:
            # plain-old atoms
            bt = '`'
            if isinstance(p1, (ast.Num, ast.Str, ast.Bytes)):
                pass
            elif isinstance(p1, str):
                if p1.startswith(bt) and p1.endswith(bt):
                    p1 = ast.Str(s=p1.strip(bt),
                                 lineno=self.lineno,
                                 col_offset=self.col)
                    p1 = xonsh_regexpath(p1, lineno=self.lineno, col=self.col)
                elif p1.startswith('$'):
                    p1 = self._envvar_by_name(p1[1:],
                                              lineno=self.lineno,
                                              col=self.col)
                else:
                    raise ValueError(p1 + " not understood")
            elif isinstance(p1, LexToken):
                p1_value = p1.value
                if p1_value == 'True':
                    p1 = ast.NameConstant(value=True,
                                          lineno=p1.lineno,
                                          col_offset=p1.lexpos)
                elif p1_value == 'False':
                    p1 = ast.NameConstant(value=False,
                                          lineno=p1.lineno,
                                          col_offset=p1.lexpos)
                elif p1_value == 'None':
                    p1 = ast.NameConstant(value=None,
                                          lineno=p1.lineno,
                                          col_offset=p1.lexpos)
                elif p1_value == '...':
                    p1 = ast.Ellipsis(lineno=p1.lineno, col_offset=p1.lexpos)
                else:
                    p1 = ast.Name(id=p1_value,
                                  ctx=ast.Load(),
                                  lineno=p1.lineno,
                                  col_offset=p1.lexpos)
            else:
                raise ValueError('{0!r} not understood'.format(p1))
            p[0] = p1
            return
        if isinstance(p1, LexToken):
            p1, p1_tok = p1.value, p1
        p2 = p[2]
        if p2 is None:
            # empty container atoms
            if p1 == '(':
                p0 = ast.Tuple(elts=[],
                               ctx=ast.Load(),
                               lineno=self.lineno,
                               col_offset=self.col)
            elif p1 == '[':
                p0 = ast.List(elts=[],
                              ctx=ast.Load(),
                              lineno=self.lineno,
                              col_offset=self.col)
            elif p1 == '{':
                p0 = ast.Dict(keys=[],
                              values=[],
                              ctx=ast.Load(),
                              lineno=self.lineno,
                              col_offset=self.col)
            else:
                assert False
        elif p1 == '(':
            # filled, possible group container tuple atoms
            if isinstance(p2, ast.AST):
                p0 = p2
                p0._lopen_lineno, p0._lopen_col = p1_tok.lineno, p1_tok.lexpos
                p0._real_tuple = True
            elif len(p2) == 1 and isinstance(p2[0], ast.AST):
                p0 = p2[0]
                p0._lopen_lineno, p0._lopen_col = p1_tok.lineno, p1_tok.lexpos
            else:
                assert False
        elif p1 == '[':
            if isinstance(p2, ast.GeneratorExp):
                p0 = ast.ListComp(elt=p2.elt,
                                  generators=p2.generators,
                                  lineno=p2.lineno,
                                  col_offset=p2.col_offset)
            else:
                if isinstance(p2, ast.Tuple):
                    if hasattr(p2, '_real_tuple') and p2._real_tuple:
                        elts = [p2]
                    else:
                        elts = p2.elts
                else:
                    elts = [p2]
                p0 = ast.List(elts=elts,
                              ctx=ast.Load(),
                              lineno=p1_tok.lineno,
                              col_offset=p1_tok.lexpos)
        elif p1 == '{':
            p0 = p2
            p0.lineno, p0.col_offset = p1_tok.lineno, p1_tok.lexpos
        elif p1.startswith('$'):
            p0 = self._dollar_rules(p)
        else:
            assert False
        p[0] = p0

    def p_string_literal(self, p):
        """string_literal : string_tok"""
        p1 = p[1]
        s = ast.literal_eval(p1.value)
        cls = ast.Bytes if p1.value.startswith('b') else ast.Str
        p[0] = cls(s=s, lineno=p1.lineno, col_offset=p1.lexpos)

    def p_string_literal_list(self, p):
        """string_literal_list : string_literal
                               | string_literal_list string_literal
        """
        if len(p) == 3:
            p[1].s += p[2].s
        p[0] = p[1]

    def p_number(self, p):
        """number : number_tok"""
        p1 = p[1]
        p[0] = ast.Num(n=ast.literal_eval(p1.value), lineno=p1.lineno,
                       col_offset=p1.lexpos)

    def p_testlist_comp(self, p):
        """testlist_comp : test_or_star_expr comp_for
                         | test_or_star_expr comma_opt
                         | test_or_star_expr comma_test_or_star_expr_list comma_opt
        """
        p1, p2 = p[1], p[2]
        if len(p) == 3:
            if p2 is None:
                # split out grouping parentheses.
                p0 = p1
            elif p2 == ',':
                p0 = ast.Tuple(elts=[p1],
                               ctx=ast.Load(),
                               lineno=p1.lineno,
                               col_offset=p1.col_offset)
            elif 'comps' in p2:
                p0 = ast.GeneratorExp(elt=p1,
                                      generators=p2['comps'],
                                      lineno=p1.lineno,
                                      col_offset=p1.col_offset)
            else:
                assert False
        elif len(p) == 4:
            p0 = ast.Tuple(elts=[p1],
                           ctx=ast.Load(),
                           lineno=p1.lineno,
                           col_offset=p1.col_offset)
            if p2 is not None:
                p0.elts.extend(p2)
            else:
                assert False
        else:
            assert False
        p[0] = p0

    def p_trailer(self, p):
        """trailer : LPAREN arglist_opt RPAREN
                   | LBRACKET subscriptlist RBRACKET
                   | PERIOD NAME
                   | DOUBLE_QUESTION
                   | QUESTION
        """
        p1 = p[1]
        p2 = p[2] if len(p) > 2 else None
        if p1 == '[':
            p0 = [p2]
        elif p1 == '(':
            p0 = [p2 or dict(args=[], keywords=[], starargs=None, kwargs=None)]
        elif p1 == '.':
            p0 = [p2]
        elif p1 == '?' or p1 == '??':
            p0 = [p1]
        else:
            assert False
        p[0] = p0

    def p_subscriptlist(self, p):
        """subscriptlist : subscript comma_subscript_list_opt comma_opt"""
        p1, p2 = p[1], p[2]
        if p2 is not None:
            p1.value = ast.Tuple(elts=[p1.value] + [x.value for x in p2],
                                 ctx=ast.Load(),
                                 lineno=p1.lineno,
                                 col_offset=p1.col_offset)
        p[0] = p1

    def p_comma_subscript(self, p):
        """comma_subscript : COMMA subscript"""
        p[0] = [p[2]]

    def p_subscript(self, p):
        """subscript : test
                     | test_opt colon_tok test_opt sliceop_opt
        """
        p1 = p[1]
        if len(p) == 2:
            p0 = ast.Index(value=p1, lineno=p1.lineno, col_offset=p1.col_offset)
        else:
            if p1 is None:
                p2 = p[2]
                lineno, col = p2.lineno, p2.lexpos
            else:
                lineno, col = p1.lineno, p1.col_offset
            p0 = ast.Slice(lower=p1, upper=p[3], step=p[4], 
                           lineno=lineno, col_offset=col)
        p[0] = p0

    def p_sliceop(self, p):
        """sliceop : COLON test_opt"""
        p[0] = p[2]

    def p_expr_or_star_expr(self, p):
        """expr_or_star_expr : expr
                             | star_expr
        """
        p[0] = p[1]

    def p_comma_expr_or_star_expr(self, p):
        """comma_expr_or_star_expr : COMMA expr_or_star_expr"""
        p[0] = [p[2]]

    def p_exprlist(self, p):
        """exprlist : expr_or_star_expr comma_expr_or_star_expr_list comma_opt
                    | expr_or_star_expr comma_opt
        """
        p1, p2 = p[1], p[2]
        p3 = p[3] if len(p) == 4 else None
        if p2 is None and p3 is None:
            p0 = [p1]
        elif p2 == ',' and p3 is None:
            p0 = [p1]
        elif p2 is not None:
            p2.insert(0, p1)
            p0 = p2
        else:
            assert False
        p[0] = p0

    def p_testlist(self, p):
        """testlist : test comma_test_list COMMA
                    | test comma_test_list
                    | test COMMA
                    | test
        """
        lenp = len(p)
        p1 = p[1]
        if lenp > 2:
            if isinstance(p1, ast.Tuple) and \
                    (hasattr(p1, '_real_tuple') and p1._real_tuple):
                lineno, col = lopen_loc(p1)
                p1 = ast.Tuple(elts=[p1],
                               ctx=ast.Load(),
                               lineno=p1.lineno,
                               col_offset=p1.col_offset)
            else:
                p1 = ensure_has_elts(p1)
            p2 = p[2] if lenp > 2 else []
            p2 = [] if p2 == ',' else p2
            p1.elts += p2
        elif lenp == 2 and isinstance(p1, ast.Tuple) and \
                           (hasattr(p1, '_real_tuple') and p1._real_tuple):
            p1.lineno, p1.col_offset = lopen_loc(p1.elts[0])
        p[0] = p1

    @docstring_by_version(
        v34="""item : test COLON test""",
        v35=\
        """item : test COLON test
                | POW expr
        """,
        )
    def p_item(self, p):
        lenp = len(p)
        if lenp == 4:
            p0 = [p[1], p[3]]
        elif lenp == 3:
            p0 = [None, p[2]]
        else:
            assert False
        p[0] = p0

    def p_comma_item(self, p):
        """comma_item : COMMA item"""
        p[0] = p[2]

    def p_dictorsetmaker(self, p):
        """dictorsetmaker : item comp_for
                          | test_or_star_expr comp_for
                          | testlist
                          | test_or_star_expr comma_opt
                          | test_or_star_expr comma_test_or_star_expr_list comma_opt
                          | test COLON testlist
                          | item comma_item_list comma_opt
                          | test COLON test comma_item_list comma_opt
        """
        p1 = p[1]
        lenp = len(p)
        if lenp == 2:
            elts = self._list_or_elts_if_not_real_tuple(p1)
            p0 = ast.Set(elts=elts, ctx=ast.Load(), lineno=self.lineno,
                         col_offset=self.col)
        elif lenp == 3:
            p2 = p[2]
            if p2 is None or p2 == ',':
                elts = self._list_or_elts_if_not_real_tuple(p1)
                p0 = ast.Set(elts=elts, ctx=ast.Load(), lineno=self.lineno,
                            col_offset=self.col)
            else:
                comps = p2.get('comps', [])
                if isinstance(p1, list) and len(p1) == 2:
                    p0 = ast.DictComp(key=p1[0], value=p1[1], generators=comps,
                                      lineno=self.lineno, col_offset=self.col)
                else:
                    p0 = ast.SetComp(elt=p1, generators=comps, lineno=self.lineno,
                                     col_offset=self.col)
        elif lenp == 4:
            p2, p3 = p[2], p[3]
            if isinstance(p1, list) and len(p1) == 2:
                cls = ast.Dict
                keys = [p1[0]]
                vals = [p1[1]]
                for k, v in zip(p2[::2], p2[1::2]):
                    keys.append(k)
                    vals.append(v)
                lineno, col = lopen_loc(p1[0] or p2[0])
                p0 = ast.Dict(keys=keys, values=vals, ctx=ast.Load(), 
                              lineno=lineno, col_offset=col)
            elif p2 == ':':
                keys = [p1]
                vals = self._list_or_elts_if_not_real_tuple(p3)
                lineno, col = lopen_loc(p1)
                p0 = ast.Dict(keys=keys, values=vals, ctx=ast.Load(), 
                              lineno=lineno, col_offset=col)
            elif isinstance(p1, ast.AST):
                elts = [p1] + p2
                p0 = ast.Set(elts=elts, ctx=ast.Load(), lineno=self.lineno,
                             col_offset=self.col)
            else:
                assert False
        elif lenp == 6:
            p4 = p[4]
            keys = [p1]
            vals = [p[3]]
            for k, v in zip(p4[::2], p4[1::2]):
                keys.append(k)
                vals.append(v)
            lineno, col = lopen_loc(p1)
            p0 = ast.Dict(keys=keys, values=vals, ctx=ast.Load(), 
                          lineno=lineno, col_offset=col)
        else:
            assert False
        p[0] = p0

    def p_classdef(self, p):
        """classdef : class_tok NAME func_call_opt COLON suite"""
        p1, p3 = p[1], p[3]
        b, kw = ([], []) if p3 is None else (p3['args'], p3['keywords'])
        c = ast.ClassDef(name=p[2],
                         bases=b,
                         keywords=kw,
                         starargs=None,
                         kwargs=None,
                         body=p[5],
                         decorator_list=[],
                         lineno=p1.lineno,
                         col_offset=p1.lexpos)
        p[0] = [c]

    def _set_arg(self, args, arg, ensure_kw=False):
        if isinstance(arg, ast.keyword):
            args['keywords'].append(arg)
        elif ensure_kw:
            if VER_MAJOR_MINOR <= VER_3_4:
                args['kwargs'] = arg
            elif VER_MAJOR_MINOR >= VER_3_5:
                args['keywords'].append(ast.keyword(arg=None, value=arg))
        else:
            args['args'].append(arg)

    #
    # arglist rule had significant changes
    #
    if VER_3_5 <= VER_MAJOR_MINOR:
        def p_arglist(self, p):
            """arglist : argument comma_opt
                       | argument comma_argument_list comma_opt
            """
            p0 = {'args': [], 'keywords': []}
            p1, p2 = p[1], p[2]
            p2 = None if p2 == ',' else p2
            self._set_arg(p0, p1)
            if p2 is not None:
                for arg in p2:
                    self._set_arg(p0, arg)
            p[0] = p0
    else:  # Python v3.4
        def p_arglist(self, p):
            """arglist : argument comma_opt
                       | argument_comma_list argument comma_opt
                       | argument_comma_list_opt TIMES test comma_argument_list_opt
                       | argument_comma_list_opt TIMES test COMMA POW test
                       | argument_comma_list_opt TIMES test comma_argument_list COMMA POW test
                       | argument_comma_list_opt POW test
            """
            lenp = len(p)
            p1, p2 = p[1], p[2]
            p0 = {'args': [], 'keywords': [], 'starargs': None, 'kwargs': None}
            if lenp == 3:
                self._set_arg(p0, p1)
            elif lenp == 4 and p2 != '**':
                for arg in p1:
                    self._set_arg(p0, arg)
                self._set_arg(p0, p2)
            elif lenp == 4 and p2 == '**':
                if p1 is not None:
                    for arg in p1:
                        self._set_arg(p0, arg)
                self._set_arg(p0, p[3], ensure_kw=True)
            elif lenp == 5:
                p0['starargs'], p4 = p[3], p[4]
                if p1 is not None:
                    for arg in p1:
                        self._set_arg(p0, arg)
                if p4 is not None:
                    for arg in p4:
                        self._set_arg(p0, arg, ensure_kw=True)
            elif lenp == 7:
                p0['starargs'] = p[3]
                if p1 is not None:
                    for arg in p1:
                        self._set_arg(p0, arg)
                self._set_arg(p0, p[6], ensure_kw=True)
            elif lenp == 8:
                kwkey = 'keywords' if VER_MAJOR_MINOR >= VER_3_5 else 'kwargs'
                p0['starargs'], p4 = p[3], p[4]
                if p1 is not None:
                    for arg in p1:
                        self._set_arg(p0, arg)
                for arg in p4:
                    self._set_arg(p0, arg, ensure_kw=True)
                self._set_arg(p0, p[7], ensure_kw=True)
            else:
                assert False
            p[0] = p0

    if VER_MAJOR_MINOR <= VER_3_4:
        def p_argument_comma(self, p):
            """argument_comma : argument COMMA"""
            p[0] = [p[1]]

    def p_comma_argument(self, p):
        """comma_argument : COMMA argument"""
        p[0] = [p[2]]

    @docstring_by_version(
        v34 = \
        """argument : test
                    | test comp_for
                    | test EQUALS test
        """,
        v35 = \
        """argument : test_or_star_expr
                    | test comp_for
                    | test EQUALS test
                    | POW test
                    | TIMES test
        """,
        )
    def p_argument(self, p):
        # v3.4 Notes
        # Really [keyword '='] test
        # The reason that keywords are test nodes instead of NAME is that using
        # NAME results in an ambiguity.
        #
        # v3.5 Notes
        # "test '=' test" is really "keyword '=' test", but we have no such token.
        # These need to be in a single rule to avoid grammar that is ambiguous
        # to our LL(1) parser. Even though 'test' includes '*expr' in star_expr,
        # we explicitly match '*' here, too, to give it proper precedence.
        # Illegal combinations and orderings are blocked in ast.c:
        # multiple (test comp_for) arguements are blocked; keyword unpackings
        # that precede iterable unpackings are blocked; etc.
        p1 = p[1]
        lenp = len(p)
        if lenp == 2:
            p0 = p1
        elif lenp == 3:
            if p1 == '**':
                p0 = ast.keyword(arg=None, value=p[2])
            elif p1 == '*':
                p0 = ast.Starred(value=p[2])
            else:
                p0 = ast.GeneratorExp(elt=p1, generators=p[2]['comps'],
                                      lineno=p1.lineno, col_offset=p1.col_offset)
        elif lenp == 4:
            p0 = ast.keyword(arg=p1.id, value=p[3])
        else:
            assert False
        p[0] = p0

    def p_comp_iter(self, p):
        """comp_iter : comp_for
                     | comp_if
        """
        p[0] = p[1]

    def p_comp_for(self, p):
        """comp_for : FOR exprlist IN or_test comp_iter_opt"""
        targs, it, p5 = p[2], p[4], p[5]
        if len(targs) == 1:
            targ = targs[0]
        else:
            targ = ensure_has_elts(targs)
        store_ctx(targ)
        comp = ast.comprehension(target=targ, iter=it, ifs=[])
        comps = [comp]
        p0 = {'comps': comps}
        if p5 is not None:
            comps += p5.get('comps', [])
            comp.ifs += p5.get('if', [])
        p[0] = p0

    def p_comp_if(self, p):
        """comp_if : IF test_nocond comp_iter_opt"""
        p2, p3 = p[2], p[3]
        p0 = {'if': [p2]}
        if p3 is not None:
            p0['comps'] = p3.get('comps', [])
        p[0] = p0

    def p_yield_expr(self, p):
        """yield_expr : yield_tok yield_arg_opt"""
        p1, p2 = p[1], p[2]
        if p2 is None:
            p0 = ast.Yield(value=p2, lineno=p1.lineno, col_offset=p1.lexpos)
        elif p2['from']:
            p0 = ast.YieldFrom(value=p2['val'],
                               lineno=p1.lineno,
                               col_offset=p1.lexpos)
        else:
            p0 = ast.Yield(value=p2['val'],
                           lineno=p1.lineno,
                           col_offset=p1.lexpos)
        p[0] = p0

    def p_yield_arg(self, p):
        """yield_arg : FROM test
                     | testlist
        """
        if len(p) == 2:
            p0 = {'from': False, 'val': p[1]}
        else:
            p0 = {'from': True, 'val': p[2]}
        p[0] = p0

    #
    # subprocess
    #

    def _dollar_rules(self, p):
        """These handle the special xonsh $ shell atoms by looking up
        in a special __xonsh_env__ dictionary injected in the __builtin__.
        """
        lenp = len(p)
        p1, p2 = p[1], p[2]
        if isinstance(p1, LexToken):
            p1, p1_tok = p1.value, p1
            lineno, col = p1_tok.lineno, p1_tok.lexpos
        else:
            lineno, col = self.lineno, self.col
        if lenp == 3:  # $NAME
            p0 = self._envvar_by_name(p2, lineno=lineno, col=col)
        elif p1 == '${':
            xenv = self._xenv(lineno=lineno, col=col)
            idx = ast.Index(value=p2)
            p0 = ast.Subscript(value=xenv,
                               slice=idx,
                               ctx=ast.Load(),
                               lineno=lineno,
                               col_offset=col)
        elif p1 == '$(':
            p0 = xonsh_call('__xonsh_subproc_captured__', p2,
                            lineno=lineno,
                            col=col)
        elif p1 == '$[':
            p0 = xonsh_call('__xonsh_subproc_uncaptured__', p2,
                            lineno=lineno,
                            col=col)
        else:
            assert False
        return p0

    def _xenv(self, lineno=lineno, col=col):
        """Creates a new xonsh env reference."""
        return ast.Name(id='__xonsh_env__',
                        ctx=ast.Load(),
                        lineno=lineno,
                        col_offset=col)

    def _envvar_getter_by_name(self, var, lineno=None, col=None):
        xenv = self._xenv(lineno=lineno, col=col)
        func = ast.Attribute(value=xenv,
                             attr='get',
                             ctx = ast.Load(),
                             lineno=lineno,
                             col_offset=col)
        return ast.Call(func=func,
                        args=[ast.Str(s=var, lineno=lineno, col_offset=col),
                              ast.Str(s='', lineno=lineno, col_offset=col)],
                        keywords=[],
                        starargs=None,
                        kwargs=None,
                        lineno=lineno,
                        col_offset=col)

    def _envvar_by_name(self, var, lineno=None, col=None):
        """Looks up a xonsh variable by name."""
        xenv = self._xenv(lineno=lineno, col=col)
        idx = ast.Index(value=ast.Str(s=var, lineno=lineno, col_offset=col))
        return ast.Subscript(value=xenv,
                             slice=idx,
                             ctx=ast.Load(),
                             lineno=lineno,
                             col_offset=col)

    def _subproc_cliargs(self, args, lineno=None, col=None):
        """Creates an expression for subprocess CLI arguments."""
        cliargs = currlist = empty_list(lineno=lineno, col=col)
        for arg in args:
            action = arg._cliarg_action
            if action == 'append':
                if currlist is None:
                    currlist = empty_list(lineno=lineno, col=col)
                    cliargs = binop(cliargs, ast.Add(), currlist,
                                    lineno=lineno,
                                    col=col)
                currlist.elts.append(arg)
            elif action == 'extend':
                cliargs = binop(cliargs, ast.Add(), arg,
                                lineno=lineno,
                                col=col)
                currlist = None
            elif action == 'splitlines':
                sl = call_split_lines(arg, lineno=lineno, col=col)
                cliargs = binop(cliargs, ast.Add(), sl, lineno=lineno, col=col)
                currlist = None
            elif action == 'ensure_list':
                x = ensure_list_from_str_or_list(arg, lineno=lineno, col=col)
                cliargs = binop(cliargs, ast.Add(), x, lineno=lineno, col=col)
                currlist = None
            else:
                raise ValueError("action not understood: " + action)
            del arg._cliarg_action
        return cliargs

    def p_pipe(self, p):
        """pipe : PIPE
                | WS PIPE
                | PIPE WS
                | WS PIPE WS
        """
        p1 = p[1]
        if len(p) > 2 and len(p1.strip()) == 0:
            p1 = p[2]
        p[0] = ast.Str(s=p1, lineno=self.lineno, col_offset=self.col)

    def p_subproc(self, p):
        """subproc : subproc_atoms
                   | subproc_atoms WS
                   | subproc AMPERSAND
                   | subproc pipe subproc_atoms
                   | subproc pipe subproc_atoms WS
        """
        lineno = self.lineno
        col = self.col
        lenp = len(p)
        p1 = p[1]
        if lenp == 2:
            p0 = [self._subproc_cliargs(p1, lineno=lineno, col=col)]
        elif p[2] == '&':
            p0 = p1 + [ast.Str(s=p[2], lineno=lineno, col_offset=col)]
        elif lenp == 3:
            p0 = [self._subproc_cliargs(p1, lineno=lineno, col=col)]
        else:
            if len(p1) > 1 and hasattr(p1[-2], 's') and p1[-2].s != '|':
                msg = 'additional redirect following non-pipe redirect'
                self._parse_error(msg, self.currloc(lineno=lineno, column=col))
            cliargs = self._subproc_cliargs(p[3], lineno=lineno, col=col)
            p0 = p1 + [p[2], cliargs]
        # return arguments list
        p[0] = p0

    def p_subproc_atoms(self, p):
        """subproc_atoms : subproc_atom
                         | subproc_atoms WS subproc_atom
        """
        p1 = p[1]
        if len(p) < 4:
            p1 = [p1]
        else:
            p1.append(p[3])
        p[0] = p1

    def p_subproc_atom(self, p):
        """subproc_atom : subproc_arg
                        | string_literal
                        | REGEXPATH
                        | DOLLAR_NAME
                        | GT
                        | LT
                        | RSHIFT
                        | IOREDIRECT
                        | AT_LPAREN test RPAREN
                        | dollar_lbrace_tok test RBRACE
                        | dollar_lparen_tok subproc RPAREN
                        | dollar_lbracket_tok subproc RBRACKET
        """
        lenp = len(p)
        p1 = p[1]
        if isinstance(p1, LexToken):
            p1, p1_tok = p1.value, p1
        if lenp == 2:
            if isinstance(p1, str):
                p0 = ast.Str(s=p1, lineno=self.lineno, col_offset=self.col)
                bt = '`'
                if p1.startswith(bt) and p1.endswith(bt):
                    p0.s = p1.strip(bt)
                    p0 = xonsh_regexpath(p0, lineno=self.lineno, col=self.col)
                    p0._cliarg_action = 'extend'
                elif '*' in p1:
                    p0 = xonsh_call('__xonsh_glob__',
                                    args=[p0],
                                    lineno=self.lineno,
                                    col=self.col)
                    p0._cliarg_action = 'extend'
                elif p1.startswith('$'):
                    p0 = self._envvar_getter_by_name(p1[1:],
                                                     lineno=self.lineno,
                                                     col=self.col)
                    p0 = xonsh_call('__xonsh_ensure_list_of_strs__', [p0],
                                    lineno=self.lineno,
                                    col=self.col)
                    p0._cliarg_action = 'extend'
                else:
                    p0 = xonsh_call('__xonsh_expand_path__', args=[p0],
                                    lineno=self.lineno, col=self.col)
                    p0._cliarg_action = 'append'
            elif isinstance(p1, ast.Str):
                p0 = xonsh_call('__xonsh_expand_path__', args=[p1],
                                lineno=self.lineno, col=self.col)
                p0._cliarg_action = 'append'
            elif isinstance(p1, ast.AST):
                p0 = p1
                p0._cliarg_action = 'append'
            else:
                assert False
        elif p1 == '@(':
            p0 = xonsh_call('__xonsh_ensure_list_of_strs__', [p[2]],
                            lineno=self.lineno,
                            col=self.col)
            p0._cliarg_action = 'extend'
        elif p1 == '${':
            lineno, col = p1_tok.lineno, p1_tok.lexpos
            xenv = self._xenv(lineno=lineno, col=col)
            func = ast.Attribute(value=xenv,
                                 attr='get',
                                 ctx = ast.Load(),
                                 lineno=lineno,
                                 col_offset=col)
            p0 = ast.Call(func=func,
                          args=[p[2],
                                ast.Str(s='',
                                        lineno=lineno,
                                        col_offset=col)],
                          keywords=[],
                          starargs=None,
                          kwargs=None,
                          lineno=lineno,
                          col_offset=col)
            p0._cliarg_action = 'append'
        elif p1 == '$(':
            p0 = xonsh_call('__xonsh_subproc_captured__',
                            args=p[2],
                            lineno=p1_tok.lineno,
                            col=p1_tok.lexpos)
            p0._cliarg_action = 'splitlines'
        elif p1 == '$[':
            p0 = xonsh_call('__xonsh_subproc_uncaptured__',
                            args=p[2],
                            lineno=p1_tok.lineno,
                            col=p1_tok.lexpos)
            p0._cliarg_action = 'splitlines'
        else:
            assert False
        p[0] = p0

    def p_subproc_arg(self, p):
        """subproc_arg : subproc_arg_part
                       | subproc_arg subproc_arg_part
        """
        # This glues the string together after parsing
        p1 = p[1]
        if len(p) == 2:
            p0 = p1
        else:
            p0 = p1 + p[2]
        p[0] = p0

    def p_subproc_arg_part(self, p):
        """subproc_arg_part : NAME
                            | TILDE
                            | PERIOD
                            | DIVIDE
                            | MINUS
                            | PLUS
                            | COLON
                            | AT
                            | EQUALS
                            | TIMES
                            | POW
                            | MOD
                            | XOR
                            | DOUBLEDIV
                            | ELLIPSIS
                            | NONE
                            | TRUE
                            | FALSE
                            | NUMBER
                            | STRING
        """
        # Many tokens cannot be part of this list, such as $, ', ", ()
        # Use a string atom instead.
        p[0] = p[1]

    #
    # Helpers
    #

    def p_empty(self, p):
        'empty : '
        p[0] = None

    def p_error(self, p):
        if p is None:
            self._parse_error('no further code', None)
        elif p.type == 'ERRORTOKEN':
            if isinstance(p.value, BaseException):
                raise p.value
            else:
                self._parse_error(p.value, self.currloc(lineno=p.lineno,
                                                        column=p.lexpos))
        else:
            msg = 'code: {0}'.format(p.value),
            self._parse_error(msg, self.currloc(lineno=p.lineno,
                                                column=p.lexpos))



# -*- coding: utf-8 -*-
"""The xonsh shell"""
import random
import builtins
from warnings import warn

from xonsh.execer import Execer
from xonsh.environ import xonshrc_context
from xonsh.tools import XonshError


def is_readline_available():
    """Checks if readline is available to import."""
    try:
        import readline
        return True
    except Exception:  # pyreadline will sometimes fail in strange ways
        return False


def is_prompt_toolkit_available():
    """Checks if prompt_toolkit is available to import."""
    try:
        import prompt_toolkit
        return True
    except ImportError:
        return False


def prompt_toolkit_version():
    """Gets the prompt toolkit version."""
    import prompt_toolkit
    return getattr(prompt_toolkit, '__version__', '<0.57')


class Shell(object):
    """Main xonsh shell.

    Initializes execution environment and decides if prompt_toolkit or
    readline version of shell should be used.
    """

    def __init__(self, ctx=None, shell_type=None, config=None, rc=None, 
                 **kwargs):
        """
        Parameters
        ----------
        ctx : Mapping, optional
            The execution context for the shell (e.g. the globals namespace).
            If none, this is computed by loading the rc files. If not None,
            this no additional context is computed and this is used 
            directly.
        shell_type : str, optional
            The shell type to start, such as 'readline', 'prompt_toolkit', 
            or 'random'.
        config : str, optional
            Path to configuration file.
        rc : list of str, optional
            Sequence of paths to run control files.
        """
        self._init_environ(ctx, config, rc)
        env = builtins.__xonsh_env__
        # pick a valid shell
        if shell_type is not None:
            env['SHELL_TYPE'] = shell_type
        shell_type = env.get('SHELL_TYPE')
        if shell_type == 'random':
            shell_type = random.choice(('readline', 'prompt_toolkit'))
        if shell_type == 'prompt_toolkit':
            if not is_prompt_toolkit_available():
                warn('prompt_toolkit is not available, using readline instead.')
                shell_type = env['SHELL_TYPE'] = 'readline'
        # actually make the shell
        if shell_type == 'prompt_toolkit':
            vptk = prompt_toolkit_version()
            minor = int(vptk.split('.')[1])
            if minor < 57 or vptk == '<0.57':  # TODO: remove in future
                msg = ('prompt-toolkit version < v0.57 and may not work as '
                       'expected. Please update.')
                warn(msg, RuntimeWarning)
            from xonsh.ptk.shell import PromptToolkitShell
            self.shell = PromptToolkitShell(execer=self.execer,
                                            ctx=self.ctx, **kwargs)
        elif shell_type == 'readline':
            from xonsh.readline_shell import ReadlineShell
            self.shell = ReadlineShell(execer=self.execer,
                                       ctx=self.ctx, **kwargs)
        else:
            raise XonshError('{} is not recognized as a shell type'.format(
                             shell_type))
        # allows history garbace colector to start running
        builtins.__xonsh_history__.gc.wait_for_shell = False

    def __getattr__(self, attr):
        """Delegates calls to appropriate shell instance."""
        return getattr(self.shell, attr)

    def _init_environ(self, ctx, config, rc):
        self.execer = Execer(config=config)
        env = builtins.__xonsh_env__
        if ctx is None:
            rc = env.get('XONSHRC') if rc is None else rc
            self.ctx = xonshrc_context(rcfiles=rc, execer=self.execer)
        else:
            self.ctx = ctx
        builtins.__xonsh_ctx__ = self.ctx
        self.ctx['__name__'] = '__main__'



# -*- coding: utf-8 -*-
"""Misc. xonsh tools.

The following implementations were forked from the IPython project:

* Copyright (c) 2008-2014, IPython Development Team
* Copyright (C) 2001-2007 Fernando Perez <fperez@colorado.edu>
* Copyright (c) 2001, Janko Hauser <jhauser@zscout.de>
* Copyright (c) 2001, Nathaniel Gray <n8gray@caltech.edu>

Implementations:

* decode()
* encode()
* cast_unicode()
* safe_hasattr()
* indent()

"""
import ctypes
import os
import re
import sys
import builtins
import platform
import traceback
import threading
import subprocess
from contextlib import contextmanager
from collections import OrderedDict, Sequence, Set
from warnings import warn

if sys.version_info[0] >= 3:
    string_types = (str, bytes)
    unicode_type = str
else:
    string_types = (str, unicode)
    unicode_type = unicode

DEFAULT_ENCODING = sys.getdefaultencoding()

ON_WINDOWS = (platform.system() == 'Windows')
ON_MAC = (platform.system() == 'Darwin')
ON_LINUX = (platform.system() == 'Linux')
ON_ARCH = (platform.linux_distribution()[0] == 'arch')
ON_POSIX = (os.name == 'posix')
IS_ROOT = ctypes.windll.shell32.IsUserAnAdmin() != 0 if ON_WINDOWS else os.getuid() == 0

VER_3_4 = (3, 4)
VER_3_5 = (3, 5)
VER_3_5_1 = (3, 5, 1)
VER_FULL = sys.version_info[:3]
VER_MAJOR_MINOR = sys.version_info[:2]
V_MAJOR_MINOR = 'v{0}{1}'.format(*sys.version_info[:2])


def docstring_by_version(**kwargs):
    """Sets a docstring by the python version."""
    doc = kwargs.get(V_MAJOR_MINOR, None)
    if V_MAJOR_MINOR is None:
        raise RuntimeError('unrecognized version ' + V_MAJOR_MINOR)
    def dec(f):
        f.__doc__ = doc
        return f
    return dec


class XonshError(Exception):
    pass


class DefaultNotGivenType(object):
    """Singleton for representing when no default value is given."""


DefaultNotGiven = DefaultNotGivenType()


def subproc_toks(line, mincol=-1, maxcol=None, lexer=None, returnline=False):
    """Excapsulates tokens in a source code line in a uncaptured
    subprocess $[] starting at a minimum column. If there are no tokens
    (ie in a comment line) this returns None.
    """
    if lexer is None:
        lexer = builtins.__xonsh_execer__.parser.lexer
    if maxcol is None:
        maxcol = len(line) + 1
    lexer.reset()
    lexer.input(line)
    toks = []
    end_offset = 0
    for tok in lexer:
        pos = tok.lexpos
        if tok.type != 'SEMI' and pos >= maxcol:
            break
        if len(toks) == 0 and tok.type in ('WS', 'INDENT'):
            continue  # handle indentation
        elif len(toks) > 0 and toks[-1].type == 'SEMI':
            if pos < maxcol and tok.type not in ('NEWLINE', 'DEDENT', 'WS'):
                toks.clear()
            else:
                break
        if pos < mincol:
            continue
        toks.append(tok)
        if tok.type == 'NEWLINE':
            break
        elif tok.type == 'DEDENT':
            # fake a newline when dedenting without a newline
            tok.type = 'NEWLINE'
            tok.value = '\n'
            tok.lineno -= 1
            tok.lexpos = len(line)
            break
    else:
        if len(toks) == 0:
            return  # handle comment lines
        if toks[-1].type == 'SEMI':
            toks.pop()
        tok = toks[-1]
        pos = tok.lexpos
        if isinstance(tok.value, string_types):
            end_offset = len(tok.value)
        else:
            el = line[pos:].split('#')[0].rstrip()
            end_offset = len(el)
    if len(toks) == 0:
        return  # handle comment lines
    beg, end = toks[0].lexpos, (toks[-1].lexpos + end_offset)
    rtn = '$[' + line[beg:end] + ']'
    if returnline:
        rtn = line[:beg] + rtn + line[end:]
    return rtn


def subexpr_from_unbalanced(expr, ltok, rtok):
    """Attempts to pull out a valid subexpression for unbalanced grouping,
    based on opening tokens, eg. '(', and closing tokens, eg. ')'.  This
    does not do full tokenization, but should be good enough for tab
    completion.
    """
    lcnt = expr.count(ltok)
    if lcnt == 0:
        return expr
    rcnt = expr.count(rtok)
    if lcnt == rcnt:
        return expr
    subexpr = expr.rsplit(ltok, 1)[-1]
    subexpr = subexpr.rsplit(',', 1)[-1]
    subexpr = subexpr.rsplit(':', 1)[-1]
    return subexpr


def decode(s, encoding=None):
    encoding = encoding or DEFAULT_ENCODING
    return s.decode(encoding, "replace")


def encode(u, encoding=None):
    encoding = encoding or DEFAULT_ENCODING
    return u.encode(encoding, "replace")


def cast_unicode(s, encoding=None):
    if isinstance(s, bytes):
        return decode(s, encoding)
    return s


def safe_hasattr(obj, attr):
    """In recent versions of Python, hasattr() only catches AttributeError.
    This catches all errors.
    """
    try:
        getattr(obj, attr)
        return True
    except Exception:  # pylint:disable=bare-except
        return False


def indent(instr, nspaces=4, ntabs=0, flatten=False):
    """Indent a string a given number of spaces or tabstops.

    indent(str,nspaces=4,ntabs=0) -> indent str by ntabs+nspaces.

    Parameters
    ----------
    instr : basestring
        The string to be indented.
    nspaces : int (default: 4)
        The number of spaces to be indented.
    ntabs : int (default: 0)
        The number of tabs to be indented.
    flatten : bool (default: False)
        Whether to scrub existing indentation.  If True, all lines will be
        aligned to the same indentation.  If False, existing indentation will
        be strictly increased.

    Returns
    -------
    outstr : string indented by ntabs and nspaces.

    """
    if instr is None:
        return
    ind = '\t' * ntabs + ' ' * nspaces
    if flatten:
        pat = re.compile(r'^\s*', re.MULTILINE)
    else:
        pat = re.compile(r'^', re.MULTILINE)
    outstr = re.sub(pat, ind, instr)
    if outstr.endswith(os.linesep + ind):
        return outstr[:-len(ind)]
    else:
        return outstr

def get_sep():
    """ Returns the appropriate filepath separator char depending on OS and
    xonsh options set
    """
    return (os.altsep if ON_WINDOWS
            and builtins.__xonsh_env__.get('FORCE_POSIX_PATHS') else
            os.sep)


TERM_COLORS = {
    # Reset
    'NO_COLOR': '\001\033[0m\002',  # Text Reset
    # Regular Colors
    'BLACK': '\001\033[0;30m\002',  # BLACK
    'RED': '\001\033[0;31m\002',  # RED
    'GREEN': '\001\033[0;32m\002',  # GREEN
    'YELLOW': '\001\033[0;33m\002',  # YELLOW
    'BLUE': '\001\033[0;34m\002',  # BLUE
    'PURPLE': '\001\033[0;35m\002',  # PURPLE
    'CYAN': '\001\033[0;36m\002',  # CYAN
    'WHITE': '\001\033[0;37m\002',  # WHITE
    # Bold
    'BOLD_BLACK': '\001\033[1;30m\002',  # BLACK
    'BOLD_RED': '\001\033[1;31m\002',  # RED
    'BOLD_GREEN': '\001\033[1;32m\002',  # GREEN
    'BOLD_YELLOW': '\001\033[1;33m\002',  # YELLOW
    'BOLD_BLUE': '\001\033[1;34m\002',  # BLUE
    'BOLD_PURPLE': '\001\033[1;35m\002',  # PURPLE
    'BOLD_CYAN': '\001\033[1;36m\002',  # CYAN
    'BOLD_WHITE': '\001\033[1;37m\002',  # WHITE
    # Underline
    'UNDERLINE_BLACK': '\001\033[4;30m\002',  # BLACK
    'UNDERLINE_RED': '\001\033[4;31m\002',  # RED
    'UNDERLINE_GREEN': '\001\033[4;32m\002',  # GREEN
    'UNDERLINE_YELLOW': '\001\033[4;33m\002',  # YELLOW
    'UNDERLINE_BLUE': '\001\033[4;34m\002',  # BLUE
    'UNDERLINE_PURPLE': '\001\033[4;35m\002',  # PURPLE
    'UNDERLINE_CYAN': '\001\033[4;36m\002',  # CYAN
    'UNDERLINE_WHITE': '\001\033[4;37m\002',  # WHITE
    # Background
    'BACKGROUND_BLACK': '\001\033[40m\002',  # BLACK
    'BACKGROUND_RED': '\001\033[41m\002',  # RED
    'BACKGROUND_GREEN': '\001\033[42m\002',  # GREEN
    'BACKGROUND_YELLOW': '\001\033[43m\002',  # YELLOW
    'BACKGROUND_BLUE': '\001\033[44m\002',  # BLUE
    'BACKGROUND_PURPLE': '\001\033[45m\002',  # PURPLE
    'BACKGROUND_CYAN': '\001\033[46m\002',  # CYAN
    'BACKGROUND_WHITE': '\001\033[47m\002',  # WHITE
    # High Intensity
    'INTENSE_BLACK': '\001\033[0;90m\002',  # BLACK
    'INTENSE_RED': '\001\033[0;91m\002',  # RED
    'INTENSE_GREEN': '\001\033[0;92m\002',  # GREEN
    'INTENSE_YELLOW': '\001\033[0;93m\002',  # YELLOW
    'INTENSE_BLUE': '\001\033[0;94m\002',  # BLUE
    'INTENSE_PURPLE': '\001\033[0;95m\002',  # PURPLE
    'INTENSE_CYAN': '\001\033[0;96m\002',  # CYAN
    'INTENSE_WHITE': '\001\033[0;97m\002',  # WHITE
    # Bold High Intensity
    'BOLD_INTENSE_BLACK': '\001\033[1;90m\002',  # BLACK
    'BOLD_INTENSE_RED': '\001\033[1;91m\002',  # RED
    'BOLD_INTENSE_GREEN': '\001\033[1;92m\002',  # GREEN
    'BOLD_INTENSE_YELLOW': '\001\033[1;93m\002',  # YELLOW
    'BOLD_INTENSE_BLUE': '\001\033[1;94m\002',  # BLUE
    'BOLD_INTENSE_PURPLE': '\001\033[1;95m\002',  # PURPLE
    'BOLD_INTENSE_CYAN': '\001\033[1;96m\002',  # CYAN
    'BOLD_INTENSE_WHITE': '\001\033[1;97m\002',  # WHITE
    # High Intensity backgrounds
    'BACKGROUND_INTENSE_BLACK': '\001\033[0;100m\002',  # BLACK
    'BACKGROUND_INTENSE_RED': '\001\033[0;101m\002',  # RED
    'BACKGROUND_INTENSE_GREEN': '\001\033[0;102m\002',  # GREEN
    'BACKGROUND_INTENSE_YELLOW': '\001\033[0;103m\002',  # YELLOW
    'BACKGROUND_INTENSE_BLUE': '\001\033[0;104m\002',  # BLUE
    'BACKGROUND_INTENSE_PURPLE': '\001\033[0;105m\002',  # PURPLE
    'BACKGROUND_INTENSE_CYAN': '\001\033[0;106m\002',  # CYAN
    'BACKGROUND_INTENSE_WHITE': '\001\033[0;107m\002',  # WHITE
}


def fallback(cond, backup):
    """Decorator for returning the object if cond is true and a backup if cond is false.
    """
    def dec(obj):
        return obj if cond else backup
    return dec


# The following redirect classes were taken directly from Python 3.5's source
# code (from the contextlib module). This can be removed when 3.5 is released,
# although redirect_stdout exists in 3.4, redirect_stderr does not.
# See the Python software license: https://docs.python.org/3/license.html
# Copyright (c) Python Software Foundation. All rights reserved.
class _RedirectStream:

    _stream = None

    def __init__(self, new_target):
        self._new_target = new_target
        # We use a list of old targets to make this CM re-entrant
        self._old_targets = []

    def __enter__(self):
        self._old_targets.append(getattr(sys, self._stream))
        setattr(sys, self._stream, self._new_target)
        return self._new_target

    def __exit__(self, exctype, excinst, exctb):
        setattr(sys, self._stream, self._old_targets.pop())


class redirect_stdout(_RedirectStream):
    """Context manager for temporarily redirecting stdout to another file::

        # How to send help() to stderr
        with redirect_stdout(sys.stderr):
            help(dir)

        # How to write help() to a file
        with open('help.txt', 'w') as f:
            with redirect_stdout(f):
                help(pow)

    Mostly for backwards compatibility.
    """
    _stream = "stdout"


class redirect_stderr(_RedirectStream):
    """Context manager for temporarily redirecting stderr to another file."""
    _stream = "stderr"


def command_not_found(cmd):
    """Uses the debian/ubuntu command-not-found utility to suggest packages for a
    command that cannot currently be found.
    """
    if not ON_LINUX:
        return ''
    elif not os.path.isfile('/usr/lib/command-not-found'):
        # utility is not on PATH
        return ''
    c = '/usr/lib/command-not-found {0}; exit 0'
    s = subprocess.check_output(c.format(cmd), universal_newlines=True,
                                stderr=subprocess.STDOUT, shell=True)
    s = '\n'.join(s.splitlines()[:-1]).strip()
    return s


def suggest_commands(cmd, env, aliases):
    """Suggests alternative commands given an environment and aliases."""
    suggest_cmds = env.get('SUGGEST_COMMANDS')
    if not suggest_cmds:
        return
    thresh = env.get('SUGGEST_THRESHOLD')
    max_sugg = env.get('SUGGEST_MAX_NUM')
    if max_sugg < 0:
        max_sugg = float('inf')

    cmd = cmd.lower()
    suggested = {}
    for a in builtins.aliases:
        if a not in suggested:
            if levenshtein(a.lower(), cmd, thresh) < thresh:
                suggested[a] = 'Alias'

    for d in filter(os.path.isdir, env.get('PATH')):
        for f in os.listdir(d):
            if f not in suggested:
                if levenshtein(f.lower(), cmd, thresh) < thresh:
                    fname = os.path.join(d, f)
                    suggested[f] = 'Command ({0})'.format(fname)
    suggested = OrderedDict(
        sorted(suggested.items(),
               key=lambda x: suggestion_sort_helper(x[0].lower(), cmd)))
    num = min(len(suggested), max_sugg)

    if num == 0:
        rtn = command_not_found(cmd)
    else:
        oneof = '' if num == 1 else 'one of '
        tips = 'Did you mean {}the following?'.format(oneof)
        items = list(suggested.popitem(False) for _ in range(num))
        length = max(len(key) for key, _ in items) + 2
        alternatives = '\n'.join('    {: <{}} {}'.format(key + ":", length, val)
                                 for key, val in items)
        rtn = '{}\n{}'.format(tips, alternatives)
        c = command_not_found(cmd)
        rtn += ('\n\n' + c) if len(c) > 0 else ''
    return rtn


def print_exception():
    """Print exceptions with/without traceback."""
    if 'XONSH_SHOW_TRACEBACK' not in builtins.__xonsh_env__:
        sys.stderr.write('xonsh: For full traceback set: '
                         '$XONSH_SHOW_TRACEBACK = True\n')
    if builtins.__xonsh_env__.get('XONSH_SHOW_TRACEBACK'):
        traceback.print_exc()
    else:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        exception_only = traceback.format_exception_only(exc_type, exc_value)
        sys.stderr.write(''.join(exception_only))


# Modified from Public Domain code, by Magnus Lie Hetland
# from http://hetland.org/coding/python/levenshtein.py
def levenshtein(a, b, max_dist=float('inf')):
    """Calculates the Levenshtein distance between a and b."""
    n, m = len(a), len(b)
    if abs(n - m) > max_dist:
        return float('inf')
    if n > m:
        # Make sure n <= m, to use O(min(n,m)) space
        a, b = b, a
        n, m = m, n
    current = range(n + 1)
    for i in range(1, m + 1):
        previous, current = current, [i] + [0] * n
        for j in range(1, n + 1):
            add, delete = previous[j] + 1, current[j - 1] + 1
            change = previous[j - 1]
            if a[j - 1] != b[i - 1]:
                change = change + 1
            current[j] = min(add, delete, change)
    return current[n]


def suggestion_sort_helper(x, y):
    """Returns a score (lower is better) for x based on how similar
    it is to y.  Used to rank suggestions."""
    x = x.lower()
    y = y.lower()
    lendiff = len(x) + len(y)
    inx = len([i for i in x if i not in y])
    iny = len([i for i in y if i not in x])
    return lendiff + inx + iny


def escape_windows_title_string(s):
    """Returns a string that is usable by the Windows cmd.exe title
    builtin.  The escaping is based on details here and emperical testing:
    http://www.robvanderwoude.com/escapechars.php
    """
    for c in '^&<>|':
        s = s.replace(c, '^' + c)
    s = s.replace('/?', '/.')
    return s


def on_main_thread():
    """Checks if we are on the main thread or not."""
    return threading.current_thread() is threading.main_thread()


@contextmanager
def swap(namespace, name, value, default=NotImplemented):
    """Swaps a current variable name in a namespace for another value, and then
    replaces it when the context is exited.
    """
    old = getattr(namespace, name, default)
    setattr(namespace, name, value)
    yield value
    if old is default:
        delattr(namespace, name)
    else:
        setattr(namespace, name, old)

#
# Validators and contervers
#


def is_int(x):
    """Tests if something is an integer"""
    return isinstance(x, int)


def is_float(x):
    """Tests if something is a float"""
    return isinstance(x, float)


def is_string(x):
    """Tests if something is a string"""
    return isinstance(x, string_types)


def always_true(x):
    """Returns True"""
    return True


def always_false(x):
    """Returns False"""
    return False


def ensure_string(x):
    """Returns a string if x is not a string, and x if it already is."""
    if isinstance(x, string_types):
        return x
    else:
        return str(x)


def is_env_path(x):
    """This tests if something is an environment path, ie a list of strings."""
    if isinstance(x, string_types):
        return False
    else:
        return (isinstance(x, Sequence) and
                all([isinstance(a, string_types) for a in x]))


def str_to_env_path(x):
    """Converts a string to an environment path, ie a list of strings,
    splitting on the OS separator.
    """
    return x.split(os.pathsep)


def env_path_to_str(x):
    """Converts an environment path to a string by joining on the OS separator."""
    return os.pathsep.join(x)


def is_bool(x):
    """Tests if something is a boolean."""
    return isinstance(x, bool)


_FALSES = frozenset(['', '0', 'n', 'f', 'no', 'none', 'false'])

def to_bool(x):
    """"Converts to a boolean in a semantically meaningful way."""
    if isinstance(x, bool):
        return x
    elif isinstance(x, string_types):
        return False if x.lower() in _FALSES else True
    else:
        return bool(x)


def bool_to_str(x):
    """Converts a bool to an empty string if False and the string '1' if True."""
    return '1' if x else ''


_BREAKS = frozenset(['b', 'break', 's', 'skip', 'q', 'quit'])

def to_bool_or_break(x):
    if isinstance(x, string_types) and x.lower() in _BREAKS:
        return 'break'
    else:
        return to_bool(x)


def ensure_int_or_slice(x):
    """Makes sure that x is list-indexable."""
    if x is None:
        return slice(None)
    elif is_int(x):
        return x
    # must have a string from here on
    if ':' in x:
        x = x.strip('[]()')
        return slice(*(int(x) if len(x) > 0 else None for x in x.split(':')))
    else:
        return int(x)


def is_string_set(x):
    """Tests if something is a set"""
    if isinstance(x, string_types):
        return False
    else:
        return (isinstance(x, Set) and
                all([isinstance(a, string_types) for a in x]))


def csv_to_set(x):
    """Convert a comma-separated list of strings to a set of strings."""
    if not x:
        return set()
    else:
        return set(x.split(','))


def set_to_csv(x):
    """Convert a set of strings to a comma-separated list of strings."""
    return ','.join(x)


def is_bool_seq(x):
    """Tests if an object is a sequence of bools."""
    return isinstance(x, Sequence) and all(map(isinstance, x, [bool]*len(x)))


def csv_to_bool_seq(x):
    """Takes a comma-separated string and converts it into a list of bools."""
    if len(x) == 0:
        return []
    return list(map(to_bool, x.split(',')))


def bool_seq_to_csv(x):
    """Converts a sequence of bools to a comma-separated string."""
    return ','.join(map(str, x))


def is_completions_display_value(x):
    return x in {'none', 'single', 'multi'}


def to_completions_display_value(x):
    x = str(x).lower()
    if x in {'none', 'false'}:
        x = 'none'
    elif x in {'multi', 'true'}:
        x = 'multi'
    elif x in {'single'}:
        x = 'single'
    else:
        warn('"{}" is not a valid value for $COMPLETIONS_DISPLAY. '.format(x) +
             'Using "multi".', RuntimeWarning)
        x = 'multi'
    return x


# history validation

_min_to_sec = lambda x: 60.0 * float(x)
_hour_to_sec = lambda x: 60.0 * _min_to_sec(x)
_day_to_sec = lambda x: 24.0 * _hour_to_sec(x)
_month_to_sec = lambda x: 30.4375 * _day_to_sec(x)
_year_to_sec = lambda x: 365.25 * _day_to_sec(x)
_kb_to_b = lambda x: 1024 * int(x)
_mb_to_b = lambda x: 1024 * _kb_to_b(x)
_gb_to_b = lambda x: 1024 * _mb_to_b(x)
_tb_to_b = lambda x: 1024 * _tb_to_b(x)

CANON_HISTORY_UNITS = frozenset(['commands', 'files', 's', 'b'])

HISTORY_UNITS = {
    '': ('commands', int),
    'c': ('commands', int),
    'cmd': ('commands', int),
    'cmds': ('commands', int),
    'command': ('commands', int),
    'commands': ('commands', int),
    'f': ('files', int),
    'files': ('files', int),
    's': ('s', float),
    'sec': ('s', float),
    'second': ('s', float),
    'seconds': ('s', float),
    'm': ('s', _min_to_sec),
    'min': ('s', _min_to_sec),
    'mins': ('s', _min_to_sec),
    'h': ('s', _hour_to_sec),
    'hr': ('s', _hour_to_sec),
    'hour': ('s', _hour_to_sec),
    'hours': ('s', _hour_to_sec),
    'd': ('s', _day_to_sec),
    'day': ('s', _day_to_sec),
    'days': ('s', _day_to_sec),
    'mon': ('s', _month_to_sec),
    'month': ('s', _month_to_sec),
    'months': ('s', _month_to_sec),
    'y': ('s', _year_to_sec),
    'yr': ('s', _year_to_sec),
    'yrs': ('s', _year_to_sec),
    'year': ('s', _year_to_sec),
    'years': ('s', _year_to_sec),
    'b': ('b', int),
    'byte': ('b', int),
    'bytes': ('b', int),
    'kb': ('b', _kb_to_b),
    'kilobyte': ('b', _kb_to_b),
    'kilobytes': ('b', _kb_to_b),
    'mb': ('b', _mb_to_b),
    'meg': ('b', _mb_to_b),
    'megs': ('b', _mb_to_b),
    'megabyte': ('b', _mb_to_b),
    'megabytes': ('b', _mb_to_b),
    'gb': ('b', _gb_to_b),
    'gig': ('b', _gb_to_b),
    'gigs': ('b', _gb_to_b),
    'gigabyte': ('b', _gb_to_b),
    'gigabytes': ('b', _gb_to_b),
    'tb': ('b', _tb_to_b),
    'terabyte': ('b', _tb_to_b),
    'terabytes': ('b', _tb_to_b),
    }
"""Maps lowercase unit names to canonical name and conversion utilities."""

def is_history_tuple(x):
    """Tests if something is a proper history value, units tuple."""
    if isinstance(x, Sequence) and len(x) == 2 and isinstance(x[0], (int, float)) \
                               and x[1].lower() in CANON_HISTORY_UNITS:
         return True
    return False


RE_HISTORY_TUPLE = re.compile('([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s*([A-Za-z]*)')

def to_history_tuple(x):
    """Converts to a canonincal history tuple."""
    if not isinstance(x, (Sequence, float, int)):
        raise ValueError('history size must be given as a sequence or number')
    if isinstance(x, str):
        m = RE_HISTORY_TUPLE.match(x.strip())
        return to_history_tuple((m.group(1), m.group(3)))
    elif isinstance(x, (float, int)):
        return to_history_tuple((x, 'commands'))
    units, converter = HISTORY_UNITS[x[1]]
    value = converter(x[0])
    return (value, units)


def history_tuple_to_str(x):
    """Converts a valid history tuple to a canonical string."""
    return '{0} {1}'.format(*x)

#
# Check pygments
#

def pygments_version():
    """Returns the Pygments version or False."""
    try:
        import pygments
        v = pygments.__version__
    except ImportError:
        v = False
    return v

#
# prompt toolkit tools
#


COLOR_CODE_SPLIT_RE = re.compile(r'(\001\033\[[\d;m]+\002)')
TERM_COLORS_REVERSED = {v: k for k, v in TERM_COLORS.items()}
COLOR_NAME_REGEX = re.compile(r'(?:(\w+)_)?(\w+)')

_PT_COLORS = {
    'BLACK': '#000000',
    'RED': '#800000',
    'GREEN': '#008000',
    'YELLOW': '#808000',
    'BLUE': '#000080',
    'PURPLE': '#800080',
    'CYAN': '#008080',
    'WHITE': '#ffffff',
    'GRAY': '#008080',
    'INTENSE_RED': '#ff1010',
    'INTENSE_GREEN': '#00ff18',
    'INTENSE_YELLOW': '#ffff00',
    'INTENSE_BLUE': '#0000d2',
    'INTENSE_PURPLE': '#ff00ff',
    'INTENSE_CYAN': '#00ffff',
    'INTENSE_GRAY': '#c0c0c0',
    None: '',
}

_PT_STYLE = {
    'BOLD': 'bold',
    'UNDERLINE': 'underline',
}

_LAST_BG_COLOR = None


def _make_style(color_name):
    """ Convert color names to pygments styles codes """
    global _LAST_BG_COLOR

    colors = _PT_COLORS.copy()
    # Extend with custom colors
    colors.update(builtins.__xonsh_env__.get('PROMPT_TOOLKIT_COLORS'))

    if color_name == 'NO_COLOR':
        _LAST_BG_COLOR = None
        return 'noinherit'

    qualifiers, name = COLOR_NAME_REGEX.match(color_name).groups()

    if name not in colors:
        qualifiers = name
        name = None

    qualifiers = qualifiers.split('_') if qualifiers else []
    is_bg_color = False
    style = []

    for qualifier in qualifiers:
        if qualifier == 'INTENSE' and name is not None:
            name = 'INTENSE_' + name
        elif qualifier == 'BACKGROUND':
            is_bg_color = True
        elif qualifier in _PT_STYLE:
            style.append(_PT_STYLE[qualifier])

    color = colors[name]
    if is_bg_color:
        _LAST_BG_COLOR = color = 'bg:' + color
    elif _LAST_BG_COLOR:
        style.append(_LAST_BG_COLOR)

    style.append(color)

    return ' '.join(style)


def format_prompt_for_prompt_toolkit(prompt):
    """Converts a prompt with color codes to a pygments style and tokens
    """
    parts = COLOR_CODE_SPLIT_RE.split(prompt)
    # ensure that parts is [colorcode, string, colorcode, string,...]
    if len(parts[0]) == 0:
        parts = parts[1:]
    else:
        parts.insert(0, '')

    if len(parts) % 2 != 0:
        parts.append('')

    strings = parts[1::2]
    token_names = [TERM_COLORS_REVERSED.get(c, 'NO_COLOR') for c in parts[::2]]
    cstyles = [_make_style(c) for c in token_names]
    return token_names, cstyles, strings


def format_color(string, remove_escapes=True):
    """Formats strings that contain xonsh.tools.TERM_COLORS values."""
    s = string.format(**TERM_COLORS)
    if remove_escapes:
        s = s.replace('\001', '').replace('\002', '')
    return s


def print_color(string, **kwargs):
    """Print strings that contain xonsh.tools.TERM_COLORS values. By default
    `sys.stdout` is used as the output stream but an alternate can be specified
    by the `file` keyword argument.
    """
    builtins.__xonsh_shell__.shell.print_color(string, **kwargs)


_RE_STRING_START = "[bBrRuU]*"
_RE_STRING_TRIPLE_DOUBLE = '"""'
_RE_STRING_TRIPLE_SINGLE = "'''"
_RE_STRING_DOUBLE = '"'
_RE_STRING_SINGLE = "'"
_STRINGS = (_RE_STRING_TRIPLE_DOUBLE,
            _RE_STRING_TRIPLE_SINGLE,
            _RE_STRING_DOUBLE,
            _RE_STRING_SINGLE)
RE_BEGIN_STRING = re.compile("(" + _RE_STRING_START +
                             '(' + "|".join(_STRINGS) +
                             '))')
"""Regular expression matching the start of a string, including quotes and
leading characters (r, b, or u)"""

RE_STRING_START = re.compile(_RE_STRING_START)
"""Regular expression matching the characters before the quotes when starting a
string (r, b, or u, case insensitive)"""

RE_STRING_CONT = {k: re.compile(v) for k,v in {
    '"': r'((\\(.|\n))|([^"\\]))*',
    "'": r"((\\(.|\n))|([^'\\]))*",
    '"""': r'((\\(.|\n))|([^"\\])|("(?!""))|\n)*',
    "'''": r"((\\(.|\n))|([^'\\])|('(?!''))|\n)*",
}.items()}
"""Dictionary mapping starting quote sequences to regular expressions that
match the contents of a string beginning with those quotes (not including the
terminating quotes)"""


def check_for_partial_string(x):
    """
    Returns the starting index (inclusive), ending index (exclusive), and
    starting quote string of the most recent Python string found in the input.

    check_for_partial_string(x) -> (startix, endix, quote)

    Parameters
    ----------
    x : str
        The string to be checked (representing a line of terminal input)

    Returns
    -------
    startix : int (or None)
        The index where the most recent Python string found started
        (inclusive), or None if no strings exist in the input

    endix : int (or None)
        The index where the most recent Python string found ended (exclusive),
        or None if no strings exist in the input OR if the input ended in the
        middle of a Python string

    quote : str (or None)
        A string containing the quote used to start the string (e.g., b", ",
        '''), or None if no string was found.
    """
    string_indices = []
    starting_quote = []
    current_index = 0
    match = re.search(RE_BEGIN_STRING, x)
    while match is not None:
        # add the start in
        start = match.start()
        quote = match.group(0)
        lenquote = len(quote)
        current_index += start
        # store the starting index of the string, as well as the
        # characters in the starting quotes (e.g., ", ', """, r", etc)
        string_indices.append(current_index)
        starting_quote.append(quote)
        # determine the string that should terminate this string
        ender = re.sub(RE_STRING_START, '', quote)
        x = x[start + lenquote:]
        current_index += lenquote
        # figure out what is inside the string
        continuer = RE_STRING_CONT[ender]
        contents = re.match(continuer, x)
        inside = contents.group(0)
        leninside = len(inside)
        current_index += contents.start() + leninside + len(ender)
        # if we are not at the end of the input string, add the ending index of
        # the string to string_indices
        if contents.end() < len(x):
            string_indices.append(current_index)
        x = x[leninside + len(ender):]
        # find the next match
        match = re.search(RE_BEGIN_STRING, x)
    numquotes = len(string_indices)
    if numquotes == 0:
        return (None, None, None)
    elif numquotes % 2:
        return (string_indices[-1], None, starting_quote[-1])
    else:
        return (string_indices[-2], string_indices[-1], starting_quote[-1])


# expandvars is a modified version of os.path.expandvars from the Python 3.5.1
# source code (root/Lib/ntpath.py, line 353)

def _is_in_env(name):
    ENV = builtins.__xonsh_env__
    return name in ENV._d or name in ENV.defaults

def _get_env_string(name):
    ENV = builtins.__xonsh_env__
    value = ENV.get(name)
    ensurer = ENV.get_ensurer(name)
    if ensurer.detype is bool_to_str:
        value = ensure_string(value)
    else:
        value = ensurer.detype(value)
    return value


def expandvars(path):
    """Expand shell variables of the forms $var, ${var} and %var%.

    Unknown variables are left unchanged."""
    ENV = builtins.__xonsh_env__
    if isinstance(path, bytes):
        path = path.decode(encoding=ENV.get('XONSH_ENCODING'),
                           errors=ENV.get('XONSH_ENCODING_ERRORS'))
    if '$' not in path and ((not ON_WINDOWS) or ('%' not in path)):
        return path
    import string
    varchars = string.ascii_letters + string.digits + '_-'
    quote = '\''
    percent = '%'
    brace = '{'
    rbrace = '}'
    dollar = '$'
    res = path[:0]
    index = 0
    pathlen = len(path)
    while index < pathlen:
        c = path[index:index+1]
        if c == quote:   # no expansion within single quotes
            path = path[index + 1:]
            pathlen = len(path)
            try:
                index = path.index(c)
                res += c + path[:index + 1]
            except ValueError:
                res += c + path
                index = pathlen - 1
        elif c == percent and ON_WINDOWS:  # variable or '%'
            if path[index + 1:index + 2] == percent:
                res += c
                index += 1
            else:
                path = path[index+1:]
                pathlen = len(path)
                try:
                    index = path.index(percent)
                except ValueError:
                    res += percent + path
                    index = pathlen - 1
                else:
                    var = path[:index]
                    if _is_in_env(var):
                        value = _get_env_string(var)
                    else:
                        value = percent + var + percent
                    res += value
        elif c == dollar:  # variable or '$$'
            if path[index + 1:index + 2] == dollar:
                res += c
                index += 1
            elif path[index + 1:index + 2] == brace:
                path = path[index+2:]
                pathlen = len(path)
                try:
                    index = path.index(rbrace)
                except ValueError:
                    res += dollar + brace + path
                    index = pathlen - 1
                else:
                    var = path[:index]
                    try:
                        var = eval(var, builtins.__xonsh_ctx__)
                        if _is_in_env(var):
                            value = _get_env_string(var)
                        else:
                            value = dollar + brace + var + rbrace
                    except:
                        value = dollar + brace + var + rbrace
                    res += value
            else:
                var = path[:0]
                index += 1
                c = path[index:index + 1]
                while c and c in varchars:
                    var += c
                    index += 1
                    c = path[index:index + 1]
                if _is_in_env(var):
                    value = _get_env_string(var)
                else:
                    value = dollar + var
                res += value
                if c:
                    index -= 1
        else:
            res += c
        index += 1
    return res

#
# File handling tools
# 

def backup_file(fname):
    """Moves an existing file to a new name that has the current time right
    before the extension.
    """
    # lazy imports
    import shutil
    from datetime import datetime
    base, ext = os.path.splitext(fname)
    newfname = base + '.' + datetime.now().isoformat() + ext
    shutil.move(fname, newfname)


def normabspath(p):
    """Retuns as normalized absolute path, namely, normcase(abspath(p))"""
    return os.path.normcase(os.path.abspath(p))



# -*- coding: utf-8 -*-
"""Job control for the xonsh shell."""
import os
import sys
import time
import signal
import builtins
from subprocess import TimeoutExpired

from xonsh.tools import ON_WINDOWS


if ON_WINDOWS:
    def _continue(obj):
        return None


    def _kill(obj):
        return obj.kill()


    def ignore_sigtstp():
        pass


    def _set_pgrp(info):
        pass

    def wait_for_active_job(signal_to_send=None):
        """
        Wait for the active job to finish, to be killed by SIGINT, or to be
        suspended by ctrl-z.
        """
        _clear_dead_jobs()
        act = builtins.__xonsh_active_job__
        if act is None:
            return
        job = builtins.__xonsh_all_jobs__[act]
        obj = job['obj']
        if job['bg']:
            return
        while obj.returncode is None:
            try:
                obj.wait(0.01)
            except TimeoutExpired:
                pass
            except KeyboardInterrupt:
                obj.kill()
        if obj.poll() is not None:
            builtins.__xonsh_active_job__ = None

else:
    def _continue(obj):
        return signal.SIGCONT


    def _kill(obj):
        os.kill(obj.pid, signal.SIGKILL)


    def ignore_sigtstp():
        signal.signal(signal.SIGTSTP, signal.SIG_IGN)


    def _set_pgrp(info):
        try:
            info['pgrp'] = os.getpgid(info['obj'].pid)
        except ProcessLookupError:
            pass


    _shell_pgrp = os.getpgrp()

    _block_when_giving = (signal.SIGTTOU, signal.SIGTTIN, signal.SIGTSTP)


    def _give_terminal_to(pgid):
        # over-simplified version of:
        #    give_terminal_to from bash 4.3 source, jobs.c, line 4030
        # this will give the terminal to the process group pgid
        if _shell_tty is not None and os.isatty(_shell_tty):
            oldmask = signal.pthread_sigmask(signal.SIG_BLOCK, _block_when_giving)
            os.tcsetpgrp(_shell_tty, pgid)
            signal.pthread_sigmask(signal.SIG_SETMASK, oldmask)


    # check for shell tty
    try:
        _shell_tty = sys.stderr.fileno()
        if os.tcgetpgrp(_shell_tty) != os.getpgid(os.getpid()):
            # we don't own it
            _shell_tty = None
    except OSError:
        _shell_tty = None


    def wait_for_active_job(signal_to_send=None):
        """
        Wait for the active job to finish, to be killed by SIGINT, or to be
        suspended by ctrl-z.
        """
        _clear_dead_jobs()
        act = builtins.__xonsh_active_job__
        if act is None:
            return
        job = builtins.__xonsh_all_jobs__[act]
        obj = job['obj']
        if job['bg'] and job['status'] == 'running':
            return
        pgrp = job['pgrp']

        # give the terminal over to the fg process
        _give_terminal_to(pgrp)
        # if necessary, send the specified signal to this process
        # (this hook  was added because vim, emacs, etc, seem to need to have
        # the terminal when they receive SIGCONT from the "fg" command)
        if signal_to_send is not None:
            if signal_to_send == signal.SIGCONT:
                job['status'] = 'running'

            os.kill(obj.pid, signal_to_send)

            if job['bg']:
                _give_terminal_to(_shell_pgrp)
                return

        _, wcode = os.waitpid(obj.pid, os.WUNTRACED)
        if os.WIFSTOPPED(wcode):
            job['bg'] = True
            job['status'] = 'stopped'
            print()  # get a newline because ^Z will have been printed
            print_one_job(act)
        elif os.WIFSIGNALED(wcode):
            print()  # get a newline because ^C will have been printed
            obj.signal = (os.WTERMSIG(wcode), os.WCOREDUMP(wcode))
            obj.returncode = None
        else:
            obj.returncode = os.WEXITSTATUS(wcode)
            obj.signal = None

        if obj.poll() is not None:
            builtins.__xonsh_active_job__ = None

        _give_terminal_to(_shell_pgrp)  # give terminal back to the shell


def _clear_dead_jobs():
    to_remove = set()
    for num, job in builtins.__xonsh_all_jobs__.items():
        obj = job['obj']
        if obj.poll() is not None:
            to_remove.add(num)
    for i in to_remove:
        del builtins.__xonsh_all_jobs__[i]
        if builtins.__xonsh_active_job__ == i:
            builtins.__xonsh_active_job__ = None
    if builtins.__xonsh_active_job__ is None:
        _reactivate_job()


def _reactivate_job():
    if len(builtins.__xonsh_all_jobs__) == 0:
        return
    builtins.__xonsh_active_job__ = max(builtins.__xonsh_all_jobs__.items(),
                                        key=lambda x: x[1]['started'])[0]



def print_one_job(num):
    """Print a line describing job number ``num``."""
    try:
        job = builtins.__xonsh_all_jobs__[num]
    except KeyError:
        return
    act = '*' if num == builtins.__xonsh_active_job__ else ' '
    status = job['status']
    cmd = [' '.join(i) if isinstance(i, list) else i for i in job['cmds']]
    cmd = ' '.join(cmd)
    pid = job['pids'][-1]
    bg = ' &' if job['bg'] else ''
    print('{}[{}] {}: {}{} ({})'.format(act, num, status, cmd, bg, pid))


def get_next_job_number():
    """Get the lowest available unique job number (for the next job created).
    """
    _clear_dead_jobs()
    i = 1
    while i in builtins.__xonsh_all_jobs__:
        i += 1
    return i


def add_job(info):
    """
    Add a new job to the jobs dictionary.
    """
    info['started'] = time.time()
    info['status'] = 'running'
    _set_pgrp(info)
    num = get_next_job_number()
    builtins.__xonsh_all_jobs__[num] = info
    builtins.__xonsh_active_job__ = num
    if info['bg']:
        print_one_job(num)


def _default_sigint_handler(num, frame):
    raise KeyboardInterrupt


def kill_all_jobs():
    """
    Send SIGKILL to all child processes (called when exiting xonsh).
    """
    _clear_dead_jobs()
    for job in builtins.__xonsh_all_jobs__.values():
        _kill(job['obj'])


def jobs(args, stdin=None):
    """
    xonsh command: jobs

    Display a list of all current jobs.
    """
    _clear_dead_jobs()
    for j in sorted(builtins.__xonsh_all_jobs__):
        print_one_job(j)
    return None, None


def fg(args, stdin=None):
    """
    xonsh command: fg

    Bring the currently active job to the foreground, or, if a single number is
    given as an argument, bring that job to the foreground.
    """
    _clear_dead_jobs()
    if len(args) == 0:
        # start active job in foreground
        act = builtins.__xonsh_active_job__
        if act is None:
            return '', 'Cannot bring nonexistent job to foreground.\n'
    elif len(args) == 1:
        try:
            act = int(args[0])
        except ValueError:
            return '', 'Invalid job: {}\n'.format(args[0])
        if act not in builtins.__xonsh_all_jobs__:
            return '', 'Invalid job: {}\n'.format(args[0])
    else:
        return '', 'fg expects 0 or 1 arguments, not {}\n'.format(len(args))
    builtins.__xonsh_active_job__ = act
    job = builtins.__xonsh_all_jobs__[act]
    job['bg'] = False
    job['status'] = 'running'
    print_one_job(act)
    wait_for_active_job(_continue(job['obj']))


def bg(args, stdin=None):
    """
    xonsh command: bg

    Resume execution of the currently active job in the background, or, if a
    single number is given as an argument, resume that job in the background.
    """
    _clear_dead_jobs()
    if len(args) == 0:
        # start active job in foreground
        act = builtins.__xonsh_active_job__
        if act is None:
            return '', 'Cannot send nonexistent job to background.\n'
    elif len(args) == 1:
        try:
            act = int(args[0])
        except ValueError:
            return '', 'Invalid job: {}\n'.format(args[0])
        if act not in builtins.__xonsh_all_jobs__:
            return '', 'Invalid job: {}\n'.format(args[0])
    else:
        return '', 'bg expects 0 or 1 arguments, not {}\n'.format(len(args))
    builtins.__xonsh_active_job__ = act
    job = builtins.__xonsh_all_jobs__[act]
    job['bg'] = True
    # When the SIGCONT is sent job['status'] is set to running.
    print_one_job(act)
    wait_for_active_job(_continue(job['obj']))



#!/usr/bin/env python
# -*- coding: ascii -*-
"""The xonsh installer."""
# Note: Do not embed any non-ASCII characters in this file until pip has been
# fixed. See https://github.com/scopatz/xonsh/issues/487.
from __future__ import print_function, unicode_literals
import os
import sys
import json
from tempfile import TemporaryDirectory
try:
    from setuptools import setup
    from setuptools.command.sdist import sdist
    from setuptools.command.install import install
    from setuptools.command.develop import develop
    HAVE_SETUPTOOLS = True
except ImportError:
    from distutils.core import setup
    from distutils.command.sdist import sdist as sdist
    from distutils.command.install import install as install
    HAVE_SETUPTOOLS = False

try:
    from jupyter_client.kernelspec import KernelSpecManager
    HAVE_JUPYTER = True
except ImportError:
    HAVE_JUPYTER = False

from xonsh import __version__ as XONSH_VERSION

TABLES = ['xonsh/lexer_table.py', 'xonsh/parser_table.py']


CONDA = ("--conda" in sys.argv)
if CONDA:
    sys.argv.remove("--conda")


def clean_tables():
    """Remove the lexer/parser modules that are dynamically created."""
    for f in TABLES:
        if os.path.isfile(f):
            os.remove(f)
            print('Remove ' + f)


def build_tables():
    """Build the lexer/parser modules."""
    print('Building lexer and parser tables.')
    sys.path.insert(0, os.path.dirname(__file__))
    from xonsh.parser import Parser
    Parser(lexer_table='lexer_table', yacc_table='parser_table',
           outputdir='xonsh')
    sys.path.pop(0)


def install_jupyter_hook(root=None):
    """Make xonsh available as a Jupyter kernel."""
    if not HAVE_JUPYTER:
        print('Could not install Jupyter kernel spec, please install '
              'Jupyter/IPython.')
        return
    spec = {"argv": [sys.executable, "-m", "xonsh.jupyter_kernel",
                     "-f", "{connection_file}"],
            "display_name": "Xonsh",
            "language": "xonsh",
            "codemirror_mode": "shell",
           }
    if CONDA:
        d = os.path.join(sys.prefix + '/share/jupyter/kernels/xonsh/')
        os.makedirs(d, exist_ok=True)
        if sys.platform == 'win32':
            # Ensure that conda-build detects the hard coded prefix
            spec['argv'][0] = spec['argv'][0].replace(os.sep, os.altsep)
        with open(os.path.join(d, 'kernel.json'), 'w') as f:
            json.dump(spec, f, sort_keys=True)
    else:
        with TemporaryDirectory() as d:
            os.chmod(d, 0o755)  # Starts off as 700, not user readable
            with open(os.path.join(d, 'kernel.json'), 'w') as f:
                json.dump(spec, f, sort_keys=True)
            print('Installing Jupyter kernel spec...')
            KernelSpecManager().install_kernel_spec(
                d, 'xonsh', user=('--user' in sys.argv), replace=True,
                prefix=root)


class xinstall(install):
    """Xonsh specialization of setuptools install class."""
    def run(self):
        clean_tables()
        build_tables()
        install_jupyter_hook(self.root if self.root else None)
        install.run(self)


class xsdist(sdist):
    """Xonsh specialization of setuptools sdist class."""
    def make_release_tree(self, basedir, files):
        clean_tables()
        build_tables()
        sdist.make_release_tree(self, basedir, files)


if HAVE_SETUPTOOLS:
    class xdevelop(develop):
        """Xonsh specialization of setuptools develop class."""
        def run(self):
            clean_tables()
            build_tables()
            develop.run(self)


def main():
    """The main entry point."""
    if sys.version_info[0] < 3:
        sys.exit('xonsh currently requires Python 3.4+')
    try:
        if '--name' not in sys.argv:
            logo_fname = os.path.join(os.path.dirname(__file__), 'logo.txt')
            with open(logo_fname, 'rb') as f:
                logo = f.read().decode('utf-8')
            print(logo)
    except UnicodeEncodeError:
        pass
    with open(os.path.join(os.path.dirname(__file__), 'README.rst'), 'r') as f:
        readme = f.read()
    skw = dict(
        name='xonsh',
        description='A general purpose, Python-ish shell',
        long_description=readme,
        license='BSD',
        version=XONSH_VERSION,
        author='Anthony Scopatz',
        maintainer='Anthony Scopatz',
        author_email='scopatz@gmail.com',
        url='https://github.com/scopatz/xonsh',
        platforms='Cross Platform',
        classifiers=['Programming Language :: Python :: 3'],
        packages=['xonsh', 'xonsh.ptk'],
        cmdclass={'install': xinstall, 'sdist': xsdist},
        )
    if HAVE_SETUPTOOLS:
        skw['setup_requires'] = ['ply']
        skw['install_requires'] = ['ply']
        skw['entry_points'] = {
            'pygments.lexers': ['xonsh = xonsh.pyghooks:XonshLexer',
                                'xonshcon = xonsh.pyghooks:XonshConsoleLexer',
                               ],
            'console_scripts': ['xonsh = xonsh.main:main'],
            }
        skw['cmdclass']['develop'] = xdevelop
    else:
        skw['scripts'] = ['scripts/xonsh'] if 'win' not in sys.platform else ['scripts/xonsh.bat'],

    setup(**skw)


if __name__ == '__main__':
    main()
